from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import str
from past.utils import old_div
from reactionary.reactionary_lib import load_eggnog_rxn_mapping_file, load_metacyc_effective_orphan_reactions, load_eggnog_annotation, mapOG2rxn_tuples, tax_dist2root, fetch_taxon_info, fetch_taxa_of_annotation, load_annot_file, eggnog_annot2rxns
from reactionary.reactionary import coerce_taxonid_to_metacyc_taxon_frame, augment_eggnog_annotation, predict_pathways
from reactionary.warpath import predict_metacyc_pathways

from camelot_frs.camelot_frs import get_kb, get_frame, get_frames, get_frame_all_parents, get_frame_all_children, frame_parent_of_frame_p
from camelot_frs.pgdb_loader import load_pgdb
from camelot_frs.pgdb_api import uniprot_links_of_reaction, potential_generic_reaction_p, get_generic_reaction_all_subs, sub_reactions_of_reaction, enzymes_of_reaction, potential_pathway_hole_p, protein_complex_p, monomers_of_complex, get_dblinks, reactions_of_protein, complexes_of_monomer

from Bio import Entrez
from ete3 import NCBITaxa
from collections import defaultdict, namedtuple
from random import shuffle
import sys, gzip, time, os, sqlite3

"""
validate.py: Compare Reactionary predictions against PGDB:
"""



'''
Interactively:
load_pgdb("/home/taltman/data/biocyc-24.0/ecoli/24.0/data")

ecoliKB = get_kb('ECOLI')




env_tuple = load_env("/home/taltman/data/biocyc-24.0/meta/24.0/data", "/home/taltman/data/nog2rxn_v5.gz", "/home/taltman/repos/reactionary/reactionary/dbs/effective-metacyc-orphan-rxns_2020-07-21", "/home/taltman/repos/reactionary/test/ecoli-24-0-annot.emapper.annotations")

taxon2tuple = fetch_taxa_of_annotation("/home/taltman/repos/reactionary/test/ecoli-24-0-annot.emapper.annotations")

metaKB = get_kb('META')

pgdb_env_list = load_pgdb_env("/home/taltman/data/biocyc-24.0/ecoli/24.0/data", 'ECOLI', env_list[1])

env_list = load_env("/home/taltman/data/biocyc-24.0/meta/24.0/data", "/home/taltman/data/nog2rxn_v5.gz", "/home/taltman/repos/reactionary/reactionary/dbs/effective-metacyc-orphan-rxns_2020-07-21", "/home/taltman/repos/reactionary/test/ecoli-24-0-annot.emapper.annotations")

start_time = time.time(); val_list = pgdb_validate(env_list[2], '511145', env_list[0], taxon2tuple, env_list[4], pgdb_env_list[0], tax_restrict=False); print(str(time.time() - start_time))


### Debugging Bsub:

bsub_env_list = load_pgdb_env("/home/taltman/data/biocyc-24.0/bsub/42/data", 'BSUB', '/home/taltman/repos/reactionary/test/bsub-24-0-annot.emapper.annotations', env_list[0], env_list[1])

## Predicted rxns:
bsub_val_list = pgdb_validate(bsub_env_list[1], '2', env_list[0], bsub_taxon2tuple, env_list[4], bsub_env_list[0], tax_restrict=False)

## Rxns MetaCyc says should be there:
meta_bsub_rxns=[ enzrxn.get_slot_values('REACTION')[0].frame_id for enzrxn in get_frame_all_children(get_frame(metaKB, 'Enzymatic-Reactions'), frame_types = 'instance' ) if 'Bacillus subtilis 168' in enzrxn.get_slot_values('ENZYME')[0].get_slot_values('SPECIES') ]

## Differences:
len(set(meta_bsub_rxns) - set(bsub_val_list[0]))

bsub_manual_rxns = { enzrxn.get_slot_values('REACTION')[0].frame_id for enzrxn in get_frame_all_children(get_frame(bsub_kb, 'Enzymatic-Reactions'), frame_types = 'instance' ) if 'BASIS-FOR-ASSIGNMENT' in enzrxn.slots and enzrxn.get_slot_values('BASIS-FOR-ASSIGNMENT')[0] == ':MANUAL' }

bsub_val_list = pgdb_validate(bsub_env_list[1], '2', env_list[0], taxon2tuple, env_list[4], bsub_env_list[0], tax_restrict=True)

len(set(bsub_manual_rxns) - set(bsub_val_list[0]))
## Still 122 rxns not found, why?

## Removing rxns not in MetaCyc, and ones that don't have a enzymatic reaction, we're left with 49:
fn_rxns  = [rxn for rxn in list(set(bsub_manual_rxns) - set(bsub_val_list[0])) if get_frame(metaKB, rxn) and 'ENZYMATIC-REACTION' in get_frame(metaKB, rxn).slots ]

## Making sure that the reactions to scrutinize are 'testable':
len(set(fn_rxns) & set(bsub_env_list[0]))
# 27 reactions
fn_rxns = list(set(fn_rxns) & set(bsub_env_list[0]))


### New workflow:

env_tuple = load_env("/home/taltman/data/biocyc-24.0/meta/24.0/data",'/home/taltman/data/')

# ECOLI:
ecoli_env_list = load_pgdb_env('/home/taltman/data/biocyc-24.0/ecoli/24.0/data', 'ECOLI', '/home/taltman/repos/reactionary/test/ecoli-24-0-annot.emapper-annotations', env_tuple.orphan_dict, env_tuple.nog2rxn_conn)

# BSUB
bsub_env_list = load_pgdb_env("/home/taltman/data/biocyc-24.0/bsub/42/data", 'BSUB', '/home/taltman/repos/reactionary/test/bsub-24-0-annot.emapper.annotations', env_tuple.orphan_dict, env_tuple.nog2rxn_conn)

'''




## Load in MetaCyc2GO:
def load_metacyc2go(metacyc2go_file):
    metacyc2go_rxns = defaultdict(int)
    go2metacyc_rxn  = defaultdict(list)
    num_mappings = 0
    with open(metacyc2go_file) as metacyc2go_fh:
        for line in metacyc2go_fh:
            if not line.startswith('!'):
                num_mappings += 1
                metacyc_token = line.split(' > ')[0]
                metacyc_rxn = metacyc_token.split(':')[1].strip()
                metacyc2go_rxns[metacyc_rxn] = 1
                go_term = line.split(' ; ')[1].strip()            
                go2metacyc_rxn[go_term].append(metacyc_rxn)
    return metacyc2go_rxns, go2metacyc_rxn

## Load in go2metacyc mappings from MetaCyc

def load_metacyc2go_inhouse(go2metacyc_inhouse_file):
    go2metacyc_rxn_inhouse = defaultdict(list)
    metacyc_rxns_inhouse = defaultdict(int)
    with open(go2metacyc_inhouse_file) as go2metacyc_inhouse_fh:
        for line in go2metacyc_inhouse_fh:
            parts = line.strip().split('\t')
            go2metacyc_rxn_inhouse[parts[0]].append(parts[1])
            metacyc_rxns_inhouse[parts[1]]
    return metacyc_rxns_inhouse, go2metacyc_rxn_inhouse





def load_known_rxn_file(known_rxn_file):
    known_rxns = defaultdict(int)
    with open(known_rxn_file) as rxns_fh:
        for line in rxns_fh:
            known_rxns[line.strip()]
    return list(known_rxns.keys())





    


### Load Reactions from EcoCyc and select ones valid for comparison:

# Spot checking, I see that there are EcoCyc reactions that are not spontaneous nor orphans, yet have no associated enzyme information:
# Examples: RXN0-286, RXN-19340 (these are bad examples, for  they are protein reactions that would not be included.
# I should consider filtering out reactions for which there is little hope of annotating.

def rxn_classes_with_no_enzymatic_reactions():
    
    rxn_classes = get_frame_all_children(get_frame(metaKB, 'Reactions'), frame_types='class')
    rxn_class_size_list = []
    
    for rxn_class in rxn_classes:
        rxn_class_instances = get_frame_all_children(rxn_class, frame_types='instance')
        catalyzed_rxns = [ rxn for rxn in rxn_class_instances if 'ENZYMATIC-REACTION' in rxn.slots ]
        if rxn_class_instances:
            rxn_class_size_list.append([rxn_class, len(rxn_class_instances), len(catalyzed_rxns), len(catalyzed_rxns)*1.0/len(rxn_class_instances)])
        else:
            rxn_class_size_list.append([rxn_class, len(rxn_class_instances), len(catalyzed_rxns), None])
    
    return rxn_class_size_list


def get_testable_rxn_list(rxn_list, metacyc_orphans_dict):
    
    testable_rxns = []
    metaKB = get_kb('META')
    rxn_class         = get_frame(metaKB, 'Chemical-Reactions')
    rxns              = [ rxn for rxn in get_frame_all_children(rxn_class) if not rxn.class_p() ]
    meta_rxns         = [ rxn.frame_id for rxn in rxns ]
    spontaneous_rxns  = [ rxn.frame_id for rxn in rxns if rxn.get_slot_values('SPONTANEOUS?') == ['T'] ]
    super_rxns        = [ rxn.frame_id for rxn in rxns if rxn.get_slot_values('REACTION-LIST') ]
    protein_substrate_rxns = []
    
    for rxn in rxns:
        for cpd in rxn.get_slot_values('LEFT') + rxn.get_slot_values('RIGHT'):
            if type(cpd) is not str and frame_parent_of_frame_p(get_frame(metaKB, 'Proteins'), cpd):
                protein_substrate_rxns.append(rxn.frame_id)
    
    testable_rxns = ( set(rxn_list) & set(meta_rxns) ) \
        - set(metacyc_orphans_dict.keys()) \
        - set(spontaneous_rxns) \
        - set(super_rxns) \
        - set(protein_substrate_rxns)
    
    return testable_rxns



## When we are working with PGDBs, we can use this:
def get_pgdb_testable_reactions(orgKB, metacyc_orphans_dict,restrict_xport_p=True):
    rxn_class         = get_frame(orgKB, 'Chemical-Reactions')
    xport_rxn_class   = get_frame(orgKB, 'Transport-Reactions')
    rxns              = [ rxn for rxn in get_frame_all_children(rxn_class) if not rxn.class_p() ]
    if restrict_xport_p:
        xport_rxns        = [ rxn for rxn in get_frame_all_children(xport_rxn_class) if not rxn.class_p() ]
    else:
        xport_rxns = []
    orphan_rxns       = [ rxn for rxn in rxns if rxn.get_slot_values('ORPHAN?') == [':YES-CONFIRMED'] ]
    spontaneous_rxns  = [ rxn for rxn in rxns if rxn.get_slot_values('SPONTANEOUS?') == ['T'] ]
    pathway_holes     = [ rxn for rxn in rxns if potential_pathway_hole_p(rxn) ]
    super_rxns        = [ rxn for rxn in rxns if rxn.get_slot_values('REACTION-LIST') ]
    instantiated_rxns = [ rxn for rxn in rxns if 'COMMENT' in rxn.slots and 'instantiat' in rxn.get_slot_values('COMMENT')[0] ]
    metacyc_orphans   = get_frames(orgKB, list(metacyc_orphans_dict.keys()))
    protein_substrate_rxns = []

    ## Why are protein substrate reactions excluded?
    ## There are issues where the generic reaction and the specific reaction classify
    ## their substrates differently. Here is a quote from Amanda
    ## Mackey in an email sent on 2018-10-25:
    ## "In this case the substrate and product in RXN0-6542 are
    ## proteins (CheA and phosphorylated CheA) whereas in 2.7.13.2-RXN
    ## the substrates and products are all classified as
    ## compounds. I’m not sure how to resolve this given that
    ## RXN0-6542 is used in several two-component pways"
    ## This is a difference between MetaCyc and EcoCyc.
    ## There is also a smaller set of these reactions than
    ## small-molecule reactions. 
    ## To simplify testing, we are skipping them for now.
    
    for rxn in rxns:
        for cpd in rxn.get_slot_values('LEFT') + rxn.get_slot_values('RIGHT'):
            if type(cpd) is not str and frame_parent_of_frame_p(get_frame(orgKB, 'Proteins'), cpd):
                protein_substrate_rxns.append(rxn)
    
    testable_rxns = \
        set(rxns) \
        - set(xport_rxns) \
        - set(orphan_rxns) \
        - set(spontaneous_rxns) \
        - set(pathway_holes) \
        - set(protein_substrate_rxns) \
        - set(super_rxns) \
        - set(instantiated_rxns) \
        - set(metacyc_orphans)
    
    
    return [ rxn.frame_id for rxn in testable_rxns ]


## We should also print out all remaining reactions that are not found in MetaCyc for whatever reason:

# load_pgdb(metacyc_dir)
#metaKB = get_kb('META')
# meta_rxn_class        = get_frame(metaKB, 'Chemical-Reactions')
# meta_rxns             = [ rxn for rxn in get_frame_all_children(rxn_class) if not rxn.class_p() ]
# unanswerable_eco_rxns = list(set(testable_rxns) - set(meta_rxns))
# print "Testable EcoCyc reactions that are not found in MetaCyc:"
# print len(unanswerable_eco_rxns)
# print '\n'.join(unanswerable_eco_rxns)


def test_reactome_predictions_against_gold_standard(metacyc_predictable_rxns, reaction_set, known_rxns, metacyc_complex_enz_reactions):
    true_positives  = 0
    num_false_negatives = 0
    num_false_positives = 0
    false_positives = []
    false_negatives = []
    
    reaction_set = set(reaction_set) & set(metacyc_predictable_rxns)
    
    for rxn in known_rxns:
        if rxn not in reaction_set:
            num_false_negatives += 1
            false_negatives.append(rxn)
        else:
            true_positives += 1
    
    for rxn in reaction_set:
        if rxn not in known_rxns:
            num_false_positives += 1
            false_positives.append(rxn)
    
    print("# Known reactions (gold standard): ", len(known_rxns))
    print("# predicted reactions: ", len(reaction_set))
    print("# Known reactions not predicted: ", len(set(known_rxns)-set(reaction_set)))
    print("# TP: ", true_positives)
    print("# FP: ", num_false_positives)
    print("# FN: ", num_false_negatives)
    precision = true_positives*1.0/len(reaction_set)
    recall = true_positives*1.0/len(known_rxns)
    f1_score = old_div(2.0*(precision*recall),(precision+recall))
    print("Precision: ", precision)
    print("Recall:    ", recall)
    print("F1 score:  ", f1_score)
    shuffle(false_negatives)
    shuffle(false_positives)
    print("False Positive Report:")
    print("# FPs that are 'complex enzyme' reactions:", len(set(false_positives) & set(metacyc_complex_enz_reactions)))
    print("  Sample false positives:")
    print("  ", false_positives[0:9])
    print("False Negative Report:")
    print("  Sample false negatives:")
    print("  ", false_negatives[0:9])
    
    return [ false_positives, false_negatives ]



def load_ecocyc_metacyc(ecocyc_dir,metacyc_dir):
    load_pgdb(ecocyc_dir)
    load_pgdb(metacyc_dir)



def gen_metacyc_complex_all_rxns():
    metaKB = get_kb('META')
    
    metacyc_complex_direct_rxns = [ enzrxn.get_slot_values('REACTION')[0] for enzrxn in get_frame_all_children(get_frame(metaKB, 'Enzymatic-Reactions'), frame_types = 'instance') if 'COMPONENTS' in enzrxn.get_slot_values('ENZYME')[0].slots and len(enzrxn.get_slot_values('ENZYME')[0].get_slot_values('COMPONENTS')) > 1 ]
    metacyc_complex_indirect_rxns = []
    metacyc_complex_all_rxns = [ rxn.frame_id for rxn in metacyc_complex_direct_rxns ]
    
    for rxn in metacyc_complex_direct_rxns:
        for sub_rxn in sub_reactions_of_reaction(rxn):
            metacyc_complex_indirect_rxns.append(sub_rxn.frame_id)
            metacyc_complex_all_rxns.append(sub_rxn.frame_id)
    
    return metacyc_complex_all_rxns


### Scripts for testing particular mapping files/approaches:
def test_metacyc2go():
    
    metaKB  = get_kb('META')
    
    metacyc2go_rxns, go2metacyc_rxn = load_metacyc2go(metacyc2go_file)
    metacyc_rxns_inhouse, go2metacyc_rxn_inhouse = load_metacyc2go_inhouse(go2metacyc_inhouse_file)
    
    ecocyc_rxns = gen_pgdb_predictable_rxns('ECOLI', orphan_file)
    
    metacyc_predictable_rxns = gen_pgdb_predictable_rxns('META', orphan_file)
    
    annot_ogs, annot_go_terms = load_ecoli_eggnog_annotation(annot_file)
    
    ## Reactions directly catalyzed by an enzymatic reaction:
    
    metacyc_complex_all_rxns = gen_metacyc_complex_all_rxns()
    ## Load the predicted rxns using the GO project's metacyc2go resource:
    eggnog_rxns = defaultdict(int)
    
    for go_term in list(annot_go_terms.keys()):
        for rxn in go2metacyc_rxn[go_term]:
            eggnog_rxns[rxn]
    
    print("Predictions using GO project's metacyc2go reaction file:")
    test_reactome_predictions_against_gold_standard(metacyc_predictable_rxns, list(eggnog_rxns.keys()), ecocyc_rxns, metacyc_complex_all_rxns)



## 2113 1004 1203 646 358 1467
## Ergo, the biggest problem is that relying solely on metacyc2go to get reaction IDs is not sufficient; 1203 reactions cannot even be mapped!
## Need to explore mapping via EC numbers as an intermediary.
## I am suspicious of how the metacyc2go mapping is created, I should investigate when able to polish this tool...

def test_new_metacyc2go():
    
    metacyc_predictable_rxns = gen_pgdb_predictable_rxns('META', orphan_file)
    ecocyc_rxns = gen_pgdb_predictable_rxns('ECOLI', orphan_file)
    metacyc_complex_all_rxns = gen_metacyc_complex_all_rxns()
    
    ## Evaluate predicted rxns using my own metacyc2go mapping resource:
    eggnog_rxns = defaultdict(int)
    for go_term in list(annot_go_terms.keys()):
        for rxn in go2metacyc_rxn_inhouse[go_term]:
            eggnog_rxns[rxn]
    
    
    print("Predictions using hand-rolled GO 2 MetaCyc reaction file:")
    test_reactome_predictions_against_gold_standard(metacyc_predictable_rxns, list(eggnog_rxns.keys()), ecocyc_rxns, metacyc_complex_all_rxns)




## tax_id: like 'TAX-511145'
def pgdb_validate(annot_ogs,
                  tax_id,
                  nog2rxn_conn,
                  taxon2tuple,
                  metacyc_predictable_rxn,
                  org_rxns,
                  tax_restrict=False,
                  per_annot_p=True,
                  annot_struct_dict={},
                  max_rxns=10,
                  enclosing_rank='superkingdom'):
    
    ## Local definitions
    ##eggnog_rxns = defaultdict(int)
    ##ogs_no_hits = defaultdict(int)
    
    #metacyc_predictable_rxns = get_pgdb_testable_reactions(get_kb('META'), orphan_dict, restrict_xport_p=False)
    #org_rxns = get_pgdb_testable_reactions(orgKB, orphan_dict)
    
    org_taxon = fetch_taxon_info([tax_id])[tax_id]
    
    # else:
    #     org_rxns = get_testable_rxn_list(org_rxn_list, orphan_dict)
    
    
    #metaKB = get_kb('META')
    
    
    ## Predicted reactions using EggNog mapping:
    prediction_list = []
    all_predicted_rxns = []
    num_ogs = len(list(annot_ogs.keys()))
    cursor = nog2rxn_conn.cursor()
    all_nog2rxn_nogs = [ row['OG'] for row in cursor.execute('select distinct(OG) from nog2rxn') ]
    num_findable_ogs = len(set(annot_ogs.keys()) & set(all_nog2rxn_nogs))
    
    print("Reachability report: ")
    print("Reachable OGs: ", num_findable_ogs)
    print("Out of # predicted OGs:", num_ogs)
    
    if per_annot_p:
        for prot_id in annot_struct_dict.keys():
            reactions,_,_,_ = eggnog_annot2rxns(annot_struct_dict[prot_id],
                                                org_taxon,
                                                nog2rxn_conn,
                                                taxon2tuple,
                                                max_rxns = max_rxns,
                                                enclosing_rank = enclosing_rank )
            all_predicted_rxns.extend(reactions)
    else:
        
        for og in list(annot_ogs.keys()):
            og_hits = 0
            
            og_reactions = mapOG2rxn(og, nog2rxn_conn, taxon2tuple, org_taxon, tax_restrict=tax_restrict)
            
            all_predicted_rxns.extend(og_reactions)
            #        for rxn in eggnog2rxns[og]:            
            #            rxn_obj = get_frame(metaKB, rxn)
            # if tax_restrict == True:
            #     if 'ENZYMATIC-REACTION' not in rxn_obj.slots or thereis_enzyme_of_taxon_p(get_frame(metaKB, tax_id), enzymes_of_reaction(rxn_obj, include_composites=True)):
            #         eggnog_rxns[rxn] += 1
            #         annot_ogs[og] += 1
            #         og_hits += 1
            #         prediction_list.append([og, rxn])
            # else:
            #     eggnog_rxns[rxn] += 1
            #     annot_ogs[og] += 1
            #     og_hits += 1
            #     prediction_list.append([og, rxn])
    
    all_predicted_rxns = list(set(all_predicted_rxns))
    
            # if og_hits == 0:
            # ogs_no_hits[og]
        
    
    
    print("Predictions using UniProt-EggNOG mapping:")
    metacyc_complex_all_rxns = gen_metacyc_complex_all_rxns()
    #fp, fn = [], []
    fp, fn = test_reactome_predictions_against_gold_standard(metacyc_predictable_rxns, all_predicted_rxns, org_rxns, metacyc_complex_all_rxns)
    return [all_predicted_rxns, fp, fn, annot_ogs]



### Debugging utils:
def is_subreaction_p(rxn_id, eggnog_rxn_mapping):
    is_subreaction_p = False
    for nog, uni_acc, rxn_id, enzrxn_id, annot, subrxn in eggnog_rxn_mapping:
        if subrxn == rxn_id:
            is_subreaction_p = True
    return is_subreaction_p

def fp_report(fp_rxn, annot_ogs, org_rxn_list, rxn2map, nog2rxn_conn):
    rxn_map_entries = rxn2map[fp_rxn]
    mapped_ogs      = list(set([ map_entry[0] for map_entry in rxn_map_entries ]))
    ogs_in_annot    = [ og for og in mapped_ogs if og in annot_ogs ]
    maps = []
    prots = set()
    rxns_from_ogs = set()
    for og in ogs_in_annot:
        for map_entry in eggnog2rxns[og]:
            maps.append(map_entry)
            prots.add(map_entry[1])
            rxns_from_ogs.add(map_entry[2])
        
    
    print("False Positive report for reaction: " + fp_rxn)
    print("# map entries: " + str(len(rxn_map_entries)))
    print("# mapped OGs: " + str(len(mapped_ogs)))
    print("# mapped OGs in annotation: " + str(len(ogs_in_annot)))
    print("# prots from ogs: " + str(len(prots)))
    print("# rxns from ogs: " + str(len(rxns_from_ogs)))
    print("# rxns from ogs in org: " + str(len(set(rxns_from_ogs) & set(org_rxn_list))))
    return prots, rxns_from_ogs, set(rxns_from_ogs) & set(org_rxn_list), mapped_ogs, ogs_in_annot

    
    

'''
I want a function that, given a fp, will:
* Find the annotation entries that might have lead to that prediction
* Display the chain of inference that might have been used
* Display the genes, proteins, enzymes, reactions for the annotation entry, as found in EcoCyc
^^^ All of these is what I'm doing manually, and it takes a long time

Another function: predicate to test whether a reaction is of a certain taxon,
based on the lineage of the enzymes that catalyze the reaction. 
For example, test whether a reaction has at least a single enzyme that is found in a bacterial genome.
'''

def trace_rxn_fn(kb, rxn_id):
    rxn = get_frame(kb, rxn_id)
    print(rxn_id, rxn.get_slot_values('EC-NUMBER')[0])
    for enzrxn in rxn.get_slot_values('ENZYMATIC-REACTION'):
        print("\t", enzrxn.frame_id, enzrxn.get_slot_values('COMMON-NAME')[0])
        for enz in monomers_of_complex(enzrxn.get_slot_values('ENZYME')[0]):
            print("\t\t\t",
                  enz.frame_id,
                  enz.get_slot_values('COMMON-NAME')[0] if 'COMMON-NAME' in enz.slots else '',
                  enz.get_slot_values('SPECIES')[0] if 'SPECIES' in enz.slots else '',
                  get_dblinks(enz, 'UNIPROT'))

                    

def trace_prot(prot_id,annot_dict,eggnog2rxns, taxon2tuple, metaKB):
    if prot_id in annot_dict:
        print(prot_id)
        for og in annot_dict[prot_id].eggnog_ids:
            print("\t" + og + ":")
            for rxn in mapOG2rxn(og, eggnog2rxns, taxon2tuple, metaKB, '', tax_restrict=False):
                print("\t\t" + rxn)
    else:
        print(prot_id + ": No match")

envData = namedtuple('envData', ['nog2rxn_conn',
                                 'orphan_dict',
                                 'metacyc_predictable_rxns'])

### Print report for a single entry in the annotation dictionary:
## This needs to handle generic-specific rxns, and sub-reactions:
def pp_annot(prot_id,
             annot_struct_dict,             
             pgdb_kb,
             org_taxon,
             nog2rxn_conn,
             taxon2tuple,
             max_rxns = 100,
             verbose_p=True,
             enclosing_rank='superkingdom'):
    
    prot_frame = get_frame(pgdb_kb, prot_id)
    pgdb_rxns  = reactions_of_protein(prot_frame)
    prot_complexes = complexes_of_monomer(prot_frame)
    cplx_rxns = [rxn for cplx in prot_complexes
                     for rxn in reactions_of_protein(cplx) ]
    prot_rxns = [ rxn.frame_id for rxn in (pgdb_rxns + cplx_rxns)]
    (pred_reactions,
     mapped_rank,
     mapped_taxon,
     prot_id,
     og) = eggnog_annot2rxns(annot_struct_dict[prot_id],
                                             org_taxon,
                                             nog2rxn_conn,
                                             taxon2tuple,
                                             max_rxns = max_rxns,
                                             enclosing_rank=enclosing_rank)

    if verbose_p:
        print('# Report for protein ' + prot_id)
        print('## Annotation:')
        print(annot_struct_dict[prot_id])
        print('')
        print('## PGDB Stats:')
        print('### Direct Reactions of Protein:')
        print('- ' + ','.join([ rxn.frame_id for rxn in pgdb_rxns ]))
        if len(prot_complexes) > 0:
            print('### Complexes of Protein:')
            for cplx in prot_complexes:
                print('--- ' + cplx.frame_id)
                cplx_rxns = reactions_of_protein(cplx)
                if len(cplx_rxns) > 0:
                    print('#### Complex Reactions')
                    print('---- ' + ','.join([rxn.frame_id for rxn in cplx_rxns]))
        print('')
        print('## Predictions')
        if(enclosing_rank != mapped_rank):
            print('### Modified taxonomic constraint:')
            print('--- enclosing rank: ' + mapped_rank)
            print('--- enclosing taxon: ' + mapped_taxon)
        print('### Basis for Predictions:')
        print('--- OG: ' + og)
        print('### Predicted Reactions:')
        print('--- ' + ','.join(pred_reactions))
    return (prot_rxns, pred_reactions)

def categorize_prots(annot_struct_dict,             
                     pgdb_kb,
                     org_taxon,
                     nog2rxn_conn,
                     taxon2tuple,
                     max_rxns = 100,
                     enclosing_rank = "order"):
    equal_prots = []
    fp_prots = []
    fn_prots = []
    fn_fp_prots = []

    for prot in annot_struct_dict.keys():
        prot_rxns, pred_rxns = pp_annot(prot,
                                        annot_struct_dict,
                                        pgdb_kb,
                                        org_taxon,
                                        nog2rxn_conn,
                                        taxon2tuple,
                                        verbose_p = False,
                                        enclosing_rank = enclosing_rank)
        fp_rxns = set(pred_rxns) - set(prot_rxns)
        fn_rxns = set(prot_rxns) - set(pred_rxns)
        if set(pred_rxns) == set(prot_rxns):
            equal_prots.append(prot)
        elif len(fp_rxns) > 0 and len(fn_rxns) > 0:
            fn_fp_prots.append(prot)
        elif len(fp_rxns) > 0:
            fp_prots.append(prot)
        elif len(fn_rxns) > 0:
            fn_prots.append(prot)

    return (equal_prots, fn_fp_prots, fp_prots, fn_prots)
    

                
## This pre-loads the reference files for Reactionary & MetaCyc:
def load_env(metacyc_dir, ref_db_path):
    
    load_pgdb(metacyc_dir)
    #eggnog2rxns = load_eggnog_rxn_mapping_file(ref_db_path)
    orphan_file = os.path.dirname(__file__) + '/dbs/effective-metacyc-orphan-rxns_2020-07-21'
    orphan_dict = load_metacyc_effective_orphan_reactions(orphan_file)
    metaKB = get_kb('META')

    ## set up the ref DB object:
    conn = sqlite3.connect(ref_db_path + '/reactionary.db')
    conn.row_factory = sqlite3.Row
    
    metacyc_predictable_rxns = get_pgdb_testable_reactions(get_kb('META'), orphan_dict, restrict_xport_p=False)
    return envData(nog2rxn_conn = conn,
                   orphan_dict  = orphan_dict,
                   metacyc_predictable_rxns = metacyc_predictable_rxns)
                   

pgdbEnvData = namedtuple('pgdbEnvData',
                         ['pgdb_testable_reactions',
                          'annot_ogs',
                          'annot_go_terms',
                          'taxon2tuple',
                          'annot_dict'])

## This loads in the reference data for a given PGDB,
## so that it doesn't need to be recomputed after many runs.
def load_pgdb_env(pgdb_dir, org_id, annot_file, orphan_dict, nog2rxn_conn):
    load_pgdb(pgdb_dir)
    pgdb_kb = get_kb(org_id)
    annot_ogs, annot_go_terms = load_eggnog_annotation(annot_file, nog2rxn_conn, annot_mode='most_specific')
    annot_dict = load_annot_file(annot_file)
    taxon2tuple = fetch_taxa_of_annotation(annot_file)
    return pgdbEnvData(pgdb_testable_reactions = get_pgdb_testable_reactions(pgdb_kb, orphan_dict, restrict_xport_p=False),
                       annot_ogs = annot_ogs,
                       annot_go_terms = annot_go_terms,
                       taxon2tuple = taxon2tuple,
                       annot_dict = annot_dict)

### Print PGDB report, comparing Reactionary pathway predictions to PGDB pathways for a select set of pathway classes:

def get_warpath_pwys(taxon_id,
                     annot_path,
                     conn,
                     enclosing_rank):

    output_dir = '/tmp'
    ncbi_taxon_tuple = fetch_taxon_info([taxon_id])[taxon_id]
    taxon2tuple = fetch_taxa_of_annotation(annot_path)
    rxn2annot = augment_eggnog_annotation(annot_path,
                                          taxon2tuple,
                                          ncbi_taxon_tuple,
                                          conn,
                                          output_dir,
                                          max_rxns = 10,
                                          enclosing_rank = enclosing_rank)
    annotated_eggnog_file = output_dir + '/' + \
                            os.path.basename(annot_path)+'.reactionary'
    pwy_pred_results, _ = predict_pathways(annotated_eggnog_file,
                                           ncbi_taxon_tuple,
                                           output_dir)
    return pwy_pred_results, rxn2annot

def warpath_vs_pgdb_roc(taxon_id, predicted_pwys, pgdb, metaKB):
    for threshold in range(0,105,5):
        f1, precision, recall, \
            tp, fp, fn = warpath_vs_pgdb_stats(taxon_id,
                                               predicted_pwys,
                                               threshold/100,
                                               pgdb,
                                               metaKB)
        print(f'{f1:.2f}\t{precision:.2f}\t{recall:.2f}\t{tp}\t{fp}\t{fn}')
              

def eval_pwy_pred_p(taxon_id, pwy_pred, threshold):
    ncbi = NCBITaxa('/home/taltman/data/taxa.sqlite')
    tax_domain_p = len(set([str(tax_id) for tax_id in ncbi.get_lineage(taxon_id)]) & set(pwy_pred[3])) > 0
    pwy_p = tax_domain_p and pwy_pred[2] >= threshold
    if pwy_pred[6] != None:
        pwy_p = pwy_p and pwy_pred[6]
    if pwy_pred[8] != None:
        pwy_p = pwy_p and pwy_pred[8]
    
    return pwy_p
    
def warpath_vs_pgdb_stats(taxon_id, predicted_pwys, threshold, pgdb, metaKB):
    pwy_class = get_frame(metaKB, 'Pathways')
    super_pwy_class = get_frame(metaKB, 'Super-Pathways')
    predictable_pwys = [ pwy.frame_id
                         for pwy in get_frame_all_children(pwy_class, frame_types='instance')
                         if super_pwy_class not in pwy.get_frame_parents() ]
    pwy_class = get_frame(pgdb, 'Pathways')
    super_pwy_class = get_frame(pgdb, 'Super-Pathways')
    ecoli_predictable_pwys = [ pwy.frame_id
                               for pwy in get_frame_all_children(pwy_class, frame_types='instance')
                               if super_pwy_class not in pwy.get_frame_parents() ]

    predicted_pwy_ids = [ pred[0] for pred in predicted_pwys if eval_pwy_pred_p(taxon_id,
                                                                                pred,
                                                                                threshold) ]

    true_positives  = set(ecoli_predictable_pwys) & set(predicted_pwy_ids)
    false_positives = set(predicted_pwy_ids) - set(ecoli_predictable_pwys)
    false_negatives = set(ecoli_predictable_pwys) - set(predicted_pwy_ids)
    precision = len(true_positives)/len(predicted_pwy_ids)
    recall    = len(true_positives)/len(ecoli_predictable_pwys)
    f1_score = 2.0*precision*recall/(precision+recall)
    return (f1_score,
            precision,
            recall,
            len(true_positives),
            len(false_positives),
            len(false_negatives))


def warpath_vs_pgdb_pwys_report(env_tuple,
                                pgdb_dir,
                                org_id,
                                annot_path,
                                rxn_taxon_id,
                                pgdb_taxon_id,
                                pwy_class_list):
    metaKB = get_kb('META')
    
    pgdb_env_tuple = load_pgdb_env(pgdb_dir,
				   org_id,
				   annot_path,
				   env_tuple.orphan_dict,
                                   env_tuple.nog2rxn_conn)
    pgdb_kb = get_kb(org_id)
    
    pgdb_val_list = pgdb_validate(pgdb_env_tuple.annot_ogs,
                                  rxn_taxon_id,
                                  env_tuple.nog2rxn_conn,
                                  pgdb_env_tuple.taxon2tuple,
                                  env_tuple.metacyc_predictable_rxns,
                                  pgdb_env_tuple.pgdb_testable_reactions,
                                  annot_struct_dict = pgdb_env_tuple.annot_dict,
                                  tax_restrict=True,
                                  per_annot_p=True)
    
    # pgdb_taxon_frame = coerce_taxonid_to_metacyc_taxon_frame(pgdb_taxon_id,
    #                                                          metaKB,
    #                                                          Entrez.email)

    ncbi_taxon_tuple = fetch_taxon_info([pgdb_taxon_id])[pgdb_taxon_id]
    
    pwys = predict_metacyc_pathways(pgdb_val_list[0], ncbi_taxon_tuple)

    pwy_classes = get_frames(metaKB, pwy_class_list)
    
    pwys_dict = { pwy_pred[0]: pwy_pred for pwy_pred in pwys}
    
    print('\n### Pathway Prediction Comparison: ###\n')
    
    for pwy_class in pwy_class_list:
        pgdb_class = get_frame(pgdb_kb, pwy_class)
        meta_class = get_frame(metaKB,  pwy_class)
        if meta_class.class_p():
            pgdb_pwys_in_class = get_frame_all_children(pgdb_class,
                                                    frame_types='instance')
            warpath_pwys_in_class = [ pwy for pwy in get_frame_all_children(meta_class, frame_types='instance') if pwy.frame_id in pwys_dict ]
        else:
            pgdb_pwys_in_class = [ pgdb_class ]
            warpath_pwys_in_class = [ meta_class ] if meta_class.frame_id in pwys_dict else []
        print("[" + meta_class.frame_id + "]", meta_class.get_slot_values('COMMON-NAME')[0] + ':')
        print('\tPGDB pathways:')
        for pgdb_pwy in pgdb_pwys_in_class:
            if pgdb_pwy != None:
                print('\t\t[' + pgdb_pwy.frame_id + '] ' + pgdb_pwy.get_slot_values('COMMON-NAME')[0])
        ##print('\t' + str(pgdb_pwys_in_class) )
        print('\tWarPath predicted pathways:')
        for warpath_pwy in warpath_pwys_in_class:
            print('\t\t[' + warpath_pwy.frame_id + '] ' + warpath_pwy.get_slot_values('COMMON-NAME')[0] + ':')
            print('\t\t\tWeighted reaction norm: ' + str(pwys_dict[warpath_pwy.frame_id][2]))
            print('\t\t\tTaxonomic domain(s): ' + str(pwys_dict[warpath_pwy.frame_id][3]))
        print('')
    
    return [pgdb_env_tuple,
            pgdb_val_list,
            pwys]


def pwy_comp_report(pwy,
                    pgdbKB,
                    predicted_rxns):
    key_rxns = pwy.get_slot_values('KEY-REACTIONS')
    key_p = ''
    print('[' + pwy.frame_id + ']', pwy.get_slot_values('COMMON-NAME')[0], sep='\t')
    print('', 'ID', 'Key Rxn?', 'PGDB', 'Reactionary', sep = '\t')
    for rxn in pwy.get_slot_values('REACTION-LIST'):
        print('',
              rxn.frame_id,
              rxn in key_rxns,
              get_frame(pgdbKB, rxn.frame_id) != None,
              rxn.frame_id in predicted_rxns,
              sep = '\t')

#### Pwy Pred Validation

## Run this first:
## rxn2annot = augment_eggnog_annotation(annot_file, taxon2tuple, ncbi_taxon_tuple, metacyc_env_tuple.nog2rxn_conn, '/tmp', enclosing_rank='genus')

def run_pwy_pred_validation(orgKB, ncbi_taxon_tuple, rxn2annot):
    ecoli_pwys = [ pwy.frame_id for pwy in get_frame_all_children(get_frame(orgKB,"Pathways"),frame_types="instance") ]
    threshold_pwys = {}
    pwy_preds = predict_metacyc_pathways(list(rxn2annot.keys()), ncbi_taxon_tuple)
    for threshold in range(00,101,5):
        pred_pwys= [ pwy_data[0] for pwy_data in pwy_preds if predict_pathways(pwy_data,threshold/100) ]
        tp = set(ecoli_pwys) & set(pred_pwys)
        tn = set(ecoli_pwys) - set(pred_pwys)
        fp = set(pred_pwys) - set(ecoli_pwys)
        precision = len(tp)/len(set(pred_pwys))
        recall = len(tp)/len(set(ecoli_pwys))
        f1_score = (2.0*precision*recall)/(precision+recall)
        print(threshold,
              f1_score,
              precision,
              recall)
              

    
### When run as a script:
if __name__ == "__main__":
    
    ### Command-line arguments:
    
    metacyc2go_file         = sys.argv[1]
    annot_file              = sys.argv[2]
    ecocyc_dir              = sys.argv[3]
    go2metacyc_inhouse_file = sys.argv[4]
    metacyc_dir             = sys.argv[5]
    nog2rxn_file            = sys.argv[6]
    orphan_file             = sys.argv[7]
    known_rxn_file          = sys.argv[8]
    tax_id                  = sys.argv[9]

    # load_ecocyc_metacyc(ecocyc_dir,
    #                     metacyc_dir)
    load_pgdb(metacyc_dir)

    eggnog2rxns = load_eggnog_rxn_mapping_file(nog2rxn_file)
    orphan_dict = load_metacyc_effective_orphan_reactions(orphan_file)
    ## vvvv fix this
    annot_ogs, annot_go_terms = load_eggnog_annotation(annot_file,eggnog2rxns, annot_mode='most_specific')
    known_rxns  = load_known_rxn_file(known_rxn_file)

    # test_uniprot_eggnog([],
    #                     annot_ogs,
    #                     'TAX-511145',
    #                     eggnog2rxns, 
    #                     orphan_dict,
    #                     orgKB=get_kb('ECOLI'))

    test_uniprot_eggnog(known_rxns,
                         annot_ogs,
                        tax_id, # 'TAX-511145'
                        eggnog2rxns, 
                        orphan_dict)
    



import os
from camelot_frs.camelot_frs import get_kb, get_frame, get_frames, get_frame_all_parents, get_frame_all_children, frame_parent_of_frame_p
from camelot_frs.pgdb_loader import load_pgdb
from camelot_frs.pgdb_api import uniprot_links_of_reaction, potential_generic_reaction_p, get_generic_reaction_all_subs, sub_reactions_of_reaction, enzymes_of_reaction, potential_pathway_hole_p, taxon_in_pathway_taxonomic_range_p
from reactionary.reactionary_lib import ncbi_taxon_id_in_lineage_p

from collections import defaultdict

## Pathway Prediction

## This MetaCyc metadata should really be in the SQLite3 db:
def load_metacyc_pwy_data():
    metacyc_pwy_file = os.path.dirname(__file__) + '/dbs/metacyc_pwy_metadata.txt'
    pwy2name = {}
    pwy2rxns = defaultdict(list)
    pwy2taxa = defaultdict(list)
    rxn2pwys = defaultdict(list)
    pwy2key_rxns = defaultdict(list)
    pwy2key_non_rxns = defaultdict(list)
    
    with open(metacyc_pwy_file) as meta_fh:
        for line in meta_fh:
            fields = line.strip().split('|')
            pwy = fields[0]
            pwy2name[pwy] = fields[1]
            if fields[4]:
                pwy2key_rxns[pwy] = fields[4].split(',')
            if fields[5]:
                pwy2key_non_rxns[pwy] = fields[5].split(',')
            ## We use [4:] to remove the 'TAX-' prefix:
            for taxon in fields[2].split(','):
                taxon_str = taxon[4:]
                pwy2taxa[pwy].append(taxon_str)
            for rxn in fields[3].split(','):
                pwy2rxns[pwy].append(rxn)
                rxn2pwys[rxn].append(pwy)                
    return pwy2name, pwy2rxns, pwy2taxa, rxn2pwys, pwy2key_rxns, pwy2key_non_rxns
            
            
# def gen_rxn_weighting_hash(kb):
#     weights = defaultdict(int)
#     rxn_class = get_frame(kb, "Reactions")
#     for rxn in get_frame_all_children(rxn_class, frame_types="instance"):
#         weights[rxn] = len(rxn.get_slot_values("IN-PATHWAY"))
#     return weights

def compute_weighted_rxn_fraction(pwy, present_rxns, pwy2rxns, rxn2pwys):
    weighted_rxns_total   = 0
    weighted_present_rxns = 0
    for rxn in pwy2rxns[pwy]:
        weighted_rxns_total += len(rxn2pwys[rxn])
        if rxn in present_rxns:
            weighted_present_rxns += len(rxn2pwys[rxn])
    return weighted_present_rxns*1.0/weighted_rxns_total


def predict_metacyc_pathways(present_rxns,org_taxon_tuple):
    possible_pwys = defaultdict(int)
    pwy_predictions = []
    (pwy2name,
     pwy2rxns,
     pwy2taxa,
     rxn2pwys,
     pwy2key_rxns,
     pwy2key_non_rxns) = load_metacyc_pwy_data()
    #rxn_weights = gen_rxn_weighting_hash(kb)
    for rxn in present_rxns:
        for pwy in rxn2pwys[rxn]:
            possible_pwys[pwy]
    for pwy in list(possible_pwys.keys()):
        super_pathway_p = False
        for rxn in pwy2rxns[pwy]:
            if rxn in pwy2name:
                super_pathway_p = True
        if not super_pathway_p and ncbi_taxon_id_in_lineage_p(pwy2taxa[pwy],
                                                              org_taxon_tuple):
            all_key_rxns_present_p = None
            key_non_rxns_satisfied_p = None
            
            if pwy in pwy2key_rxns:
                all_key_rxns_present_p = len(set(pwy2key_rxns[pwy])
                                             & set(present_rxns)) == len(pwy2key_rxns[pwy])
            else:
                all_key_rxns_present_p = None
            
            ## "Satisfied" means that the logical expression is true, and thus
            ## the pathway should not be predicted:
            if pwy in pwy2key_non_rxns:
                key_non_rxns = pwy2key_non_rxns[pwy]
                if key_non_rxns[0] == 'AND':
                    ## AND logic: if #all# rxns listed are present, then the condition is satisfied:
                    key_non_rxns_satisfied_p = len(set(key_non_rxns[1:]) - set(present_rxns)) == 0
                else:
                    ## OR logic: if *any* of the rxns listed are present, then the condition is satisifed:
                    key_non_rxns_satisfied_p = len(set(key_non_rxns[1:]) & set(present_rxns)) > 0
            else:
                key_non_rxns = []
                key_non_rxns_satsified_p = None
            
            pwy_predictions.append([pwy,
                                    pwy2name[pwy],
                                    compute_weighted_rxn_fraction(pwy, 
                                                                  present_rxns,
                                                                  pwy2rxns,
                                                                  rxn2pwys),
                                    pwy2taxa[pwy],
                                    list(set(pwy2rxns[pwy]) & set(present_rxns)),
                                    list(set(pwy2rxns[pwy]) - set(present_rxns)),
                                    all_key_rxns_present_p,
                                    pwy2key_rxns[pwy],
                                    key_non_rxns_satisfied_p,
                                    key_non_rxns
                                    
            ])
    return pwy_predictions

def _predict_pathway(pred_data,threshold):
    if pred_data[7] == ['']:
        key_rxns_p = True
    else:
        key_rxns_p = pred_data[6]
    
    if pred_data[9] == ['']:
        key_non_rxns_p = True
    else:
        key_non_rxns_p = pred_data[8]
    
    return pred_data[2] >= threshold \
        and key_rxns_p \
        and key_non_rxns_p
            



from __future__ import print_function
from builtins import str
import gzip, time, pickle, shutil
from collections import defaultdict, namedtuple
import urllib.parse
import urllib.request
import requests
from contextlib import closing
from Bio import Entrez, SwissProt
from camelot_frs.pgdb_loader import load_pgdb
from camelot_frs.camelot_frs import KB, get_kb, get_frame, get_frame_all_children
from camelot_frs.pgdb_api    import potential_generic_reaction_p, get_generic_reaction_all_subs, sub_reactions_of_reaction, uniprot_links_of_reaction, monomers_of_complex, get_dblinks

'''
gen_mapping_file.py: Generate new reference files for Reactionary

This code should only need to be run once per MetaCyc release by the maintainer, and never by end-users.

Steps:
1. Obtain the latest versions of the flat-files for MetaCyc, EcoCyc, and BsubCyc
2. Update Reactionary and/or Camelot to make sure it can still work with the PGDBs without any bugs.
3. Run: load_pgdb('/path/to/metacyc/data')
4. Run create_uniprot2rxn_file
5. Run map_eggnog_rxn_mapping
6. Run print_eggnog_rxn_mapping to save results to file.
7. Run analyze_metacyc_rxns to see if effective orphan enzymes have changed.
'''

## Email address for Entrez:
Entrez.email = 'reactionary.entrez@altmananalytics.com'


#metaKB = get_kb('META')

## We want to have these mapping files self-document how they came upon the linkages, to assist with debugging.
## Instead of mapping between a UniProt accession and a reaction, there will be some keywords included, examples:
## ['RXN-12345', 'ENZRXN-1', 'direct']
## ['RXN-54321', 'ENZRXN-1', 'sub-reaction of RXN-12345']
## ['RXN-00001', 'ENZRXN-2', 'generic-reaction of RXN-99']

def create_uniprot2rxn_file(metacyc2rhea, rhea2swissprot, specific_rxn_p=True):
    uniprot2rxn = defaultdict(list)
    metaKB = get_kb('META')
    i = 0
    all_maps = []

    ## TODO: filter out sub-reactions, they are processed by the helper function.
    for enzrxn in get_frame_all_children(get_frame(metaKB, 'Enzymatic-Reactions'), frame_types = 'instance'):
        i += 1
        if (i % 100) == 0:
            print("Processed " + str(i) + " enzymatic reactions.")
        print( enzrxn.frame_id)
        for accession, rxn_relation in create_uniprot_rxn_map(enzrxn, metacyc2rhea, rhea2swissprot, specific_rxn_p=specific_rxn_p):
            uniprot2rxn[accession].append(rxn_relation)
    return uniprot2rxn

def get_metacyc_uniprot_accs(metaKB):
    accs = []
    bad_prots = [get_frame(metaKB, 'CPLX0-231'),
                 get_frame(metaKB, 'ABC-23-CPLX'),
                 get_frame(metaKB, 'ACETOLACTSYNII-CPLX')]
    print('Bad proteins in MetaCyc 24.0, check to see if issue fixed!', bad_prots)
    for enzrxn in get_frame_all_children(get_frame(metaKB, 'Enzymatic-Reactions'), frame_types = 'instance'):
        direct_enz = enzrxn.get_slot_values('ENZYME')[0]
        if direct_enz not in bad_prots:
            for accession in [ get_dblinks(monomer, 'UNIPROT') for monomer in monomers_of_complex(direct_enz)]:
                if accession:
                    accs.append(accession[0])
    return list(set(accs))

    
#TODO: Add argument for taking hash table of mappings from uniprot IDs to rhea, EC_num, and Uniprot IDs...
def create_uniprot_rxn_map(enzrxn, metacyc2rhea, rhea2swissprot, specific_rxn_p=True):
    relations = []
    direct_rxn = enzrxn.get_slot_values('REACTION')[0]
    direct_enz = enzrxn.get_slot_values('ENZYME')[0]
    rxns     = [[direct_rxn.frame_id,enzrxn.frame_id,'direct','']]

    if specific_rxn_p:
        if potential_generic_reaction_p(direct_rxn):
            for specific_rxn in get_generic_reaction_all_subs(direct_rxn):
                rxns.append([direct_rxn.frame_id,enzrxn.frame_id,'generic-specific',specific_rxn.frame_id])
    
    ## I really should filter out sub-reactions that are spontaneous or orphan'ed, but I'm punting for now,
    ## Since there shouldn't be any in theory (enzyme assigned to "composite" reaction could cover all components)
    for sub_rxn in sub_reactions_of_reaction(direct_rxn):
        rxns.append([direct_rxn.frame_id,enzrxn.frame_id,'composite-sub',sub_rxn.frame_id])
    
    ## The following logic takes the links from the direct reaction, 
    ## and shares them with all of its' "instance reactions":

    ### Some Rhea reporting:
    #for accession in [ get_dblinks(monomer, 'UNIPROT') for monomer in monomers_of_complex(direct_enz)]:
    #     print("Test:", accession)
    #     for (rhea_id, ec_num) in uniprot2rhea[accession[0]]:
    #         print("  Rhea: ", rhea_id, ec_num)
    #         for swissprot_acc in rhea2swissprot[rhea_id]:
    #             print("    SwissProt:", swissprot_acc)
    # print("  MetaCyc:", direct_rxn.frame_id, direct_rxn.get_slot_values('EC-NUMBER'))

    ## First, we add links to UniProt via Rhea:
    #if direct_rxn in metacyc2rhea:
    for rhea_id in metacyc2rhea[direct_rxn.frame_id]:
        for sprot in rhea2swissprot[rhea_id]:
            for rxn_relation in rxns:
                relations.append([sprot, rxn_relation + [rhea_id]])
                
    ## If we cannot map to UniProt via Rhea, we fall back to trying with
    try:
        ## 
        for accession in [ get_dblinks(monomer, 'UNIPROT') for monomer in monomers_of_complex(direct_enz)]:
            for rxn_relation in rxns:
                if accession:
                    relations.append([accession[0], rxn_relation + [None]])
                #uniprot2rxn[accession].append(rxn_relation)
        
        ## Here I should do a separate loop, where we go over UniProt
        ## accessions linked indirectly via Rhea reactions.
        ## Add another column at end to clarify the "original" uniprot ID, and the Rhea ID
    except:
        print("Unable to find uniprot links for reaction " + direct_rxn.frame_id)
    
    return relations

# def print_uniprot2rxn(install_path, uniprot2rxn_dict):
#     with open(install_path + '/dbs/uniprot2rxn.txt', "w") as out:
#         for acc in uniprot2rxn_dict.keys():
#             for 
#             print >>out, "\t".join([acc] + uniprot2rxn

def analyze_uniprot2rxn(metaKB, uniprot2rxn):
    enzrxns = get_frame_all_children(get_frame(metaKB,'Enzymatic-Reactions'), frame_types='instance')
    num_rxns_w_enz = len(set([enzrxn.get_slot_values('REACTION')[0] for enzrxn in enzrxns]))
    num_uniprot2rxn_rxns = len(set([relation[0] for relation_list in list(uniprot2rxn.values()) for relation in relation_list ]))
    print("Number of MetaCyc Reactions with enzymes: " + str(num_rxns_w_enz))
    print("Number of MetaCyc Reactions in uniprot2rxn: " + str(num_uniprot2rxn_rxns))

## If everything looks good using analyze_uniprot2rxn above, go ahead and "backup" the
## uniprot2rxn datastructure, as it takes over an hour to prepare.
def pickle_uniprot2rxn(uniprot2rxn,filename):
    with open(filename,'wb') as output_fh:
        pickle.dump(uniprot2rxn, output_fh, protocol=pickle.HIGHEST_PROTOCOL)

#### UniProt Queries ####

'''
We need to perform some programmatic queries of UniProt to perform the following mappings:

1. Get all UniProt accessions from proteins which catalyze reactions in MetaCyc
2. Fetch structured data for these entries from UniProt using API
3. For all entries with Rhea accessions, perform queries to pull in these additional proteins.
4. Upload total set of accessions, and download FASTA file for all accessions:

'''

## Call like:
## uniprot2rxn = create_uniprot2rxn_file(specific_rxn_p=False)
## fetch_uniprot_data(list(uniprot2rxn.keys()), 'foo@example.com')
## Took ~7 tries before it succeeded, be patient...

def fetch_uniprot_data(accs, email_address):
    with open("metacyc_uniprot_ids.txt","w") as uni_acc_fh:
        for acc in accs:
            print(acc, file=uni_acc_fh)
    
    url = "https://www.uniprot.org/uploadlists/"
    headers = {
        'User-Agent': 'Reactionary 0.1',
        'From': email_address
    }
    
    params = { 'from': 'ACC+ID',
               'to': 'ACC',
               'format': 'txt'
    }
    
    files = {'file': open('metacyc_uniprot_ids.txt')}
    
    request = requests.post(url,
                            files=files,
                            data = params,
                            headers=headers)
    
    with open('metacyc_uniprot_data.txt','w') as uni_data_fh:
        print(request.text[:-1], file=uni_data_fh)
    
    uniprot_records = []
    with open('metacyc_uniprot_data.txt') as uni_data_fh:
        for record in SwissProt.parse(uni_data_fh):
            uniprot_records.append(record)
    
    return uniprot_records

def get_record_rhea_ids(record):
    acc = record.accessions[0]
    rhea_list = []
    for comment in record.comments:            
        if comment.startswith('CATALYTIC') and 'Rhea:RHEA' in comment:
            rhea_id = comment.split(';')[1].split(',')[0].strip().split(':')[2]
            # ec_num = None
            # if "EC=" in comment:
            #     for comment_part in comment.split(';'):
            #         if "EC=" in comment_part:
            #             ec_num = comment_part.strip().strip('EC=')
            # rhea_list.append((rhea_id, ec_num))
            rhea_list.append(rhea_id)
    return rhea_list

def make_uniprot2rhea_dict(uniprot_dat_file):
    uniprot2rhea = defaultdict(list)
    rhea2uniprot = defaultdict(list)
    with open(uniprot_dat_file) as uniprot_fh:        
        for record in SwissProt.parse(uniprot_fh):
            acc = record.accessions[0]
            rhea_ids = get_record_rhea_ids(record)
            uniprot2rhea[acc] = rhea_ids
            for rhea_id in rhea_ids:
                rhea2uniprot[rhea_id].append(acc)
    return uniprot2rhea, rhea2uniprot
                

def query_uniprot_by_rhea(rhea_id, email_address):
    headers = {
        'User-Agent': 'Reactionary 0.1',
        'From': email_address
    }
    
    rhea_query_url = 'https://www.uniprot.org/uniprot/?query=annotation:(type:"catalytic activity" rhea:' + rhea_id + ')&format=txt'
    rhea_raw_prots = requests.get(rhea_query_url, headers=headers)
    
    with open("temp_rhea.txt","w") as rhea_data_fh:
        print(rhea_raw_prots.text[:-1], file=rhea_data_fh)
    
    rhea_records = []    
    with open('temp_rhea.txt') as rhea_data_fh:
        for record in SwissProt.parse(rhea_data_fh):
            rhea_records.append(record)
    return rhea_records


def download_curr_swissprot():
    url = 'ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/uniprot_sprot.dat.gz'
    
    with closing(urllib.request.urlopen(url)) as handle:
        with open('/tmp/uniprot_sprot.dat.gz', "wb") as sprot_fh:
            shutil.copyfileobj(handle, sprot_fh)

# def make_rhea2uniprot_dict(uniprot2rhea_dict, email_address):
#     rhea2uniprot = defaultdict(list)
#     for acc in list(uniprot2rhea_dict.keys())[0:5]:
#         print(acc)
#         for (rhea_id, _) in uniprot2rhea_dict[acc]:
#             rhea_records = query_uniprot_by_rhea(rhea_id, email_address)
#             for rhea_record in rhea_records:
#                 rhea2uniprot[rhea_id].append(rhea_record.accessions[0])
#     return rhea2uniprot

def load_swissprot(swissprot_file):
    
    with open(swissprot_file) as sprot_fh:
        records = [ record for record in SwissProt.parse(sprot_fh)]
    return records

def map_metacyc2uniprot(rhea2metacyc_file, rhea2uniprot_sprot):
    metacyc_rhea_ids = defaultdict(int)
    uniprot_ids = defaultdict(int)
    metacyc2rhea = defaultdict(set)
    rhea2swissprot = defaultdict(set)
    
    line_num = 0
    with open(rhea2metacyc_file) as rhea2meta_fh:
        for line in rhea2meta_fh:
            line_num += 1
            if line_num > 1:
                fields = line.strip().split('\t')
                metacyc_rhea_ids[fields[0]]
                metacyc_rhea_ids[fields[2]]
                metacyc2rhea[fields[3]].add(fields[0])
                metacyc2rhea[fields[3]].add(fields[2])
    
    line_num = 0
    with open(rhea2uniprot_sprot) as rhea2sprot_fh:
        for line in rhea2sprot_fh:
            line_num += 1
            if line_num > 1:
                fields = line.strip().split('\t')
                rhea2swissprot[fields[0]].add(fields[3])
                rhea2swissprot[fields[2]].add(fields[3])
                if fields[0] in metacyc_rhea_ids or fields[2] in metacyc_rhea_ids:
                    uniprot_ids[fields[3]]
    
    return list(uniprot_ids.keys()), metacyc2rhea, rhea2swissprot


def print_uniprot_record_as_fasta(fasta_fh, record):
    acc = record.accessions[0]
    print('>'+acc, file=fasta_fh)
    print(record.sequence, file=fasta_fh)

def filter_sprot_by_acc(accs,
                        swissprot_file,
                        fasta_filename):
    sprot_accs = []
    accs_dict = { acc:1 for acc in accs }
    with open(swissprot_file) as sprot_fh:
        with open(fasta_filename,"w") as fasta_fh:
            records = SwissProt.parse(sprot_fh)
            for record in records:
                acc = record.accessions[0]
                if acc in accs_dict:
                    sprot_accs.append(acc)
                    print_uniprot_record_as_fasta(fasta_fh,
                                                  record)
    return list(set(accs)-set(sprot_accs))

## We need a way to easily fetch TrEMBL sequences from MetaCyc:
def generate_uniprot_fasta(swissprot_file,
                           fasta_filename,
                           uniprot2rhea,
                           metaKB):
    rhea2metacyc_uniprot = defaultdict(list)
    for (acc, rhea_list) in uniprot2rhea.items():
        for (rhea_id, ec_num) in rhea_list:
            rhea2metacyc_uniprot[rhea_id].append(acc)
    rhea2swissprot = defaultdict(list)
    uniprot_ids = get_metacyc_uniprot_accs(metaKB)
    sprot_accs = defaultdict(int)
    metacyc_accs_not_sprot = []
    with open(swissprot_file) as sprot_fh:
        with open(fasta_filename,"w") as fasta_fh:
            records = SwissProt.parse(sprot_fh)
            for record in records:
                acc = record.accessions[0]
                sprot_accs[acc]
                rhea_match_p = False
                for (rhea_id, ec_num) in get_record_rhea_ids(record):
                    if rhea_id in rhea2metacyc_uniprot:
                        ## We only collect rhea mappings relevant to MetaCyc:
                        rhea2swissprot[rhea_id].append(acc)
                        rhea_match_p = True
                if acc in uniprot_ids:
                    print_uniprot_record_as_fasta(fasta_fh,
                                                  record)
                else:
                    if rhea_match_p:                        
                        print_uniprot_record_as_fasta(fasta_fh,
                                                      record)
    non_sprot_accs = [ acc for acc in uniprot_ids if acc not in sprot_accs]
    return rhea2swissprot, non_sprot_accs

## Finally, we have to fetch the TrEMBL sequences, and print them out:
## Like the above step fetching the MetaCyc entries, this is prone to failure due to the UniProt web service being down. Will fail with message like: "Failed to find ID in first line." Manually inspect the 'metacyc_uniprot_data.txt' file to see full error message from UniProt.

def print_non_sprot_records(non_sprot_accs, fasta_file, email_address):
    non_sprot_records = fetch_uniprot_data(non_sprot_accs,
                                           email_address)
    with open(fasta_file, 'w') as out_fh:
        for record in non_sprot_records:
            print_uniprot_record_as_fasta(out_fh, record)

'''
TODO:
- Create script to merge the two FASTA files, instead of doing so manually.
'''
            
#### Process EggNOG output: 

def map_eggnog_rxn_mapping(annot_path, uniprot2rxn):
    #uniprot2rxn = create_uniprot2rxn_file()
    eggnog_rxn_mapping = []
    eggnog_uniprot_mapping = [] 
    uniprot_acs_w_eggnog_annot = defaultdict(int)
    metacyc_uniprots_w_annot  = defaultdict(int)
    metacyc_uniprots_w_eggnog = defaultdict(int)
    line_num = 0
    num_mappings = 0
    
    ### Add mappings coming from manually running emapper over UniProt proteins without EggNOG entries:
    
    with open(annot_path) as annot_fh: ##install_path + '/dbs/uniprot-metacyc-all/annot.emapper.annotations') as annot_fh:
        for line in annot_fh:
            if not line.startswith('#'):
                line_num += 1
                if (line_num % 100000) == 0:
                    print("Processed " + str(line_num) + " EggNOG annotations.")            
                fields = line.strip().split('\t')
                uniprot_ac = fields[0] #.split('|')[1]
                eggnog_ids = fields[18].split(',')
                if uniprot_ac in uniprot2rxn:
                    metacyc_uniprots_w_annot[uniprot_ac]
                if len(eggnog_ids) > 0:
                    uniprot_acs_w_eggnog_annot[uniprot_ac]
                    for raw_eggnog_id in eggnog_ids:
                        if len(raw_eggnog_id) > 0:
                            og_parts = raw_eggnog_id.split('@')
                            eggnog_id = og_parts[0]
                            taxon_id  = og_parts[1]
                        if uniprot_ac in uniprot2rxn:
                            metacyc_uniprots_w_eggnog[uniprot_ac]
                            for rxn_relation in uniprot2rxn[uniprot_ac]:
                                eggnog_rxn_mapping.append([raw_eggnog_id, taxon_id, uniprot_ac] + rxn_relation)
                                num_mappings += 1
                                if (num_mappings % 100000) == 0:
                                    print("  Added " + str(num_mappings) + " EggNOG-MetaCyc mappings.")
                        else:
                            eggnog_uniprot_mapping.append([raw_eggnog_id, taxon_id, uniprot_ac])
    
    ## Print some stats on the mapping work:
    print("Mapping Stats:")
    print("Number of UniProt accession numbers harvested from MetaCyc: ", len(list(uniprot2rxn.keys())))
    print("Number of MetaCyc UniProt accession numbers in annot file: ", len(metacyc_uniprots_w_annot))
    print("Number of MetaCyc UniProt accession numbers with EggNOG mapping(s): ", len(metacyc_uniprots_w_eggnog))
    print("Number of UniProt accessions with EggNOG annotation: ", len(list(uniprot_acs_w_eggnog_annot.keys())))
    print("Number of UniProt accessions without direct mapping: ", len(set([rxn for [rxn, _, _] in eggnog_uniprot_mapping])))
    
    return [eggnog_rxn_mapping, eggnog_uniprot_mapping, list(set(uniprot2rxn.keys()) - set(metacyc_uniprots_w_annot.keys()))]



def print_eggnog_rxn_mapping(eggnog_rxn_mapping,filename):
    with open(filename,'w') as output_fh:
        for entry in eggnog_rxn_mapping:
            str_entry = map(str, entry)
            print('\t'.join(list(str_entry)), file=output_fh)
    

# go_enzyme_instance_rxns = {}
# for rxn in go_enzyme_rxns.keys():
#     print rxn
#     instance_rxns = get_generic_reaction_all_subs(rxn)
#     if instance_rxns:
#         for inst_rxn in instance_rxns:
#             go_enzyme_instance_rxns[inst_rxn] = 1


## We use this function to print a working list of "effective orphans":
def analyze_metacyc_rxns(metaKB):
    
    all_rxns          = get_frame_all_children(get_frame(metaKB, 'Reactions'), frame_types='instance')
    direct_catalyzed_rxns    = [ rxn for rxn in all_rxns if 'ENZYMATIC-REACTION' in rxn.slots ]
    orphan_rxns       = [ rxn for rxn in all_rxns if rxn.get_slot_values('ORPHAN?') == [':YES-CONFIRMED'] ]
    spontaneous_rxns  = [ rxn for rxn in all_rxns if rxn.get_slot_values('SPONTANEOUS?') == ['T'] ]
    catalyzed_sub_rxns = defaultdict(int)
    catalyzed_instance_rxns = defaultdict(int)
    
    for rxn in direct_catalyzed_rxns:
        for sub_rxn in sub_reactions_of_reaction(rxn):
            catalyzed_sub_rxns[sub_rxn]
    
    for rxn in direct_catalyzed_rxns:
        print(rxn)
        for instance_rxn in get_generic_reaction_all_subs(rxn):
            catalyzed_instance_rxns[instance_rxn]
    
    uncertain_status = set(all_rxns) - set(direct_catalyzed_rxns) - set(orphan_rxns) - set(spontaneous_rxns) - set(catalyzed_sub_rxns.keys()) - set(catalyzed_instance_rxns.keys())
    
    return [all_rxns, direct_catalyzed_rxns, orphan_rxns, spontaneous_rxns, list(catalyzed_sub_rxns.keys()), list(catalyzed_instance_rxns.keys()), uncertain_status]



#### Code no longer used:

# enzymes = [enzrxn.get_slot_values('ENZYME')[0] for enzrxn in get_frame_all_children(get_frame(ecoliKB, 'Enzymatic-Reactions'), frame_types = 'instance')]
# go_enzymes = [enzyme for enzyme in enzymes if len(enzyme.get_slot_values('GO-TERMS')) > 0]


# metaKB = get_kb('META')
# go_enzyme_rxns = {}
# for enzrxn in get_frame_all_children(get_frame(metaKB, 'Enzymatic-Reactions'), frame_types = 'instance'):
#     print_enzrxn_go_rxn_relations(enzrxn)

def print_enzrxn_go_rxn_relations(enzrxn):
    enzyme     = enzrxn.get_slot_values('ENZYME')[0]
    direct_rxn = enzrxn.get_slot_values('REACTION')[0]
    enzymes    = [enzyme]
    rxns       = [direct_rxn]
    
    if protein_complex_p(enzyme):
        enzymes += monomers_of_complex(enzyme)
    
    if potential_generic_reaction_p(direct_rxn):
        rxns += get_generic_reaction_all_subs(direct_rxn)
        
    for monomer in enzymes:
        for rxn in rxns:
            for term in monomer.get_slot_values('GO-TERMS'):
                print('\t'.join([term.frame_id, rxn.frame_id]))


## This currently gets us 10619 out of 10866 UniProt proteins mapped to EggNOG
## The rest are oddball entries, like viral proteins. Punting for now.
def map_eggnog_rxn_mapping_original(install_path, uniprot2rxn):
    #uniprot2rxn = create_uniprot2rxn_file()
    eggnog_rxn_mapping = []
    uniprot_acs_w_eggnog_annot = defaultdict(int)
    uniprot_acs_w_eggnog_annot_from_tab = defaultdict(int)
    uniprot_acs_w_eggnog_annot_from_emapper = defaultdict(int)
    
    ### Add mappings coming from UniProt:
    with gzip.open(install_path + '/dbs/uniprot-metacyc-w-tax-info.tsv.gz','rt') as swissprot_eggnog:
        for line in swissprot_eggnog:
            if not line.startswith('Entry'):
                fields = line.strip().split('\t')
                ## This is necessary because some rows are "deleted" entries.
                if len(fields) > 6:
                    print(fields[0])
                    eggnog_ids = fields[6].split(';')
                    if len(eggnog_ids) > 1:
                        uniprot_acs_w_eggnog_annot[fields[0]]
                        uniprot_acs_w_eggnog_annot_from_tab[fields[0]]
                    for eggnog_id in eggnog_ids:
                        if len(eggnog_id) > 0:
                            for rxn_relation in uniprot2rxn[fields[0]]:
                                eggnog_rxn_mapping.append([eggnog_id, fields[0]] + rxn_relation)
    
    ### Add mappings coming from manually running emapper over UniProt proteins without EggNOG entries:
    
    ## Archaea:
    for domain in ['archaea', 'bacteria', 'eukaryota']:
        with open(install_path + '/dbs/uniprot-metacyc-no-eggnog-' + domain + '/annot.emapper.annotations') as annot_fh:
            for line in annot_fh:
                if not line.startswith('#'):
                    fields = line.strip().split('\t')
                    print(fields[0])
                    uniprot_ac = fields[0].split('|')[1]
                    eggnog_ids = fields[9].split(',')
                    if len(eggnog_ids) > 1:
                        uniprot_acs_w_eggnog_annot[uniprot_ac]
                        uniprot_acs_w_eggnog_annot_from_emapper[uniprot_ac]
                    for raw_eggnog_id in eggnog_ids:
                        if len(raw_eggnog_id) > 0:
                            og_parts = raw_eggnog_id.split('@')
                            if len(og_parts) > 1 and (og_parts[0].startswith('KOG') or og_parts[0].startswith('COG')):
                                eggnog_id = og_parts[0]
                            elif len(og_parts) > 1 and not og_parts[0].startswith('KOG'):
                                eggnog_id = 'ENOG41' + og_parts[0]
                            else:
                                eggnog_id = og_parts[0]
                            for rxn_relation in uniprot2rxn[uniprot_ac]:
                                eggnog_rxn_mapping.append([eggnog_id, uniprot_ac] + rxn_relation)
                        
    
    ## Print some stats on the mapping work:
    print("Mapping Stats:")
    print("Number of UniProt accession numbers harvested from MetaCyc: ", len(list(uniprot2rxn.keys())))
    print("Number of UniProt accessions with EggNOG annotation: ", len(list(uniprot_acs_w_eggnog_annot.keys())))
    print("   Number of UniProt accessions with EggNOG annotation from tab-delimited file: ", len(list(uniprot_acs_w_eggnog_annot_from_tab.keys())))
    print("   Number of UniProt accessions with EggNOG annotation from emapper output: ", len(list(uniprot_acs_w_eggnog_annot_from_emapper.keys())))
        
    return eggnog_rxn_mapping


## print out MetaCyc pwy report:
def print_metacyc_pwy_report(metaKB,output_file):
    pwys = get_frame_all_children(get_frame(metaKB, 'Pathways'), frame_types='instance')
    with open(output_file,'w') as output_fh:
        for pwy in pwys:
            if 'ENGINEERED?' not in pwy.slots:
                if 'KEY-NON-REACTIONS' in pwy.slots:
                    key_non_reactions = pwy.get_slot_values('KEY-NON-REACTIONS')[0].strip('()').split(' ')
                else:
                    key_non_reactions = ''
                print('|'.join((pwy.frame_id,
                                pwy.get_slot_values('COMMON-NAME')[0],
                                ','.join([taxon.frame_id for taxon in pwy.get_slot_values('TAXONOMIC-RANGE')]),
                                ','.join([rxn.frame_id for rxn in pwy.get_slot_values('REACTION-LIST')]),
                                ','.join([rxn.frame_id for rxn in pwy.get_slot_values('KEY-REACTIONS')]),
                                ','.join(key_non_reactions)
                )),
                      file=output_fh)



from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from camelot_frs.pgdb_loader import load_pgdb
from camelot_frs.camelot_frs import get_kb, get_frame, get_frames, frame_subsumes_or_equal_p
import reactionary.reactionary_lib
from reactionary.reactionary_lib import load_eggnog_rxn_mapping_file, find_og_in_annot_line, map_filter_taxon2domain, eggnog_annot2rxns, parse_annot_line
from reactionary.warpath import predict_metacyc_pathways
from ete3 import NCBITaxa
from Bio import Entrez
from collections import defaultdict
from pathlib import Path
import sys, gzip, os, urllib.request, subprocess, textwrap

eggnog_mapper_OG_field = 18
rxn_field              = 22

## Reaction annotation data file path:
#ref_db_old_url = "https://ndownloader.figshare.com/files/25440413?private_link=9f94fa9056cdbb4f0b83"


def fetch_ref_db(download_url, local_db_path):

    ## First, fetch the compressed SQL database:
    print("Fetching the compressed database...", file=sys.stderr)
    local_filename, headers = urllib.request.urlretrieve(download_url)    
    print("... done!", file=sys.stderr)
    
    ## Then, attempt to decompress the SQL and rebuild the SQLite3 DB:
    print("Decompressing and rebuilding the compressed database...", file=sys.stderr)
    output = subprocess.check_output('gunzip < ' + local_filename + ' | sqlite3 ' + local_db_path, shell=True)
    print("... done!", file=sys.stderr)


def add_refs2ref_db(new_refs_path,
                    nog2rxn_conn):
    new_refs = []
    cursor = nog2rxn_conn.cursor()
    with open(new_refs_path) as refs_fh:
        for line in refs_fh:
            entry = line.strip().split("\t")
            cursor.execute("INSERT INTO nog2rxn VALUES (?, '', '', ?, '', 'USER', '', '')", entry)
    nog2rxn_conn.commit()

def augment_eggnog_annotation(eggnog_annot_file,
                              annot_taxon2tuple,
                              ncbi_taxon,
                              nog2rxn_conn,
                              output_dir,
                              max_rxns=10,
                              enclosing_rank='superkingdom'):
    
    line_num = 0
    rxn2annot = defaultdict(list)
    max_rxn_prot_annots = []
    #eggnog2rxns  = load_eggnog_rxn_mapping_file(nog2rxn_file)

    ## Print augmented EggNOG-Mapper output file:
    with open(output_dir + '/' + os.path.basename(eggnog_annot_file) + '.reactionary', "w") as aug_annot_fh:
        with open(eggnog_annot_file) as annot_fh:
            for line in annot_fh:
                line_num += 1
                reactions = []
                ##print "Line: " + str(line_num)
                if line.startswith('#query_name'):
                    print(line.strip() + '\tMetaCyc_Reaction', file=aug_annot_fh)
                elif line.startswith('#'):
                    print(line.strip(), file=aug_annot_fh)
                else:
                    fields   = line.strip().split('\t')
                    curr_annot = parse_annot_line(line)
                    (reaction_tuples,
                     mapped_rank,
                     mapped_taxon,
                     prot_id,
                     og              ) = eggnog_annot2rxns(curr_annot,
                                                           ncbi_taxon,
                                                           nog2rxn_conn,
                                                           annot_taxon2tuple,
                                                           enclosing_rank = enclosing_rank)
                    rxn_list = list(set([ rxn_tuple[3] for rxn_tuple in reaction_tuples ]))
                    if len(reaction_tuples) <= max_rxns:
                        for rxn in rxn_list:
                            rxn2annot[rxn].append([reaction_tuples, mapped_rank, mapped_taxon, prot_id, og])
                        fields.append(",".join(rxn_list))
                        print("\t".join(fields), file=aug_annot_fh)
                    else:
                        max_rxn_prot_annots.append([reaction_tuples,
                                                    mapped_rank,
                                                    mapped_taxon,
                                                    prot_id,
                                                    og])
    
    ## Print reaction report:
    print_reaction_report(output_dir, rxn2annot)
    
    ## Print annotations which have exceeded the max number of reactions predicted:
    print_multi_reaction_protein_report(output_dir, max_rxn_prot_annots)
    
    return rxn2annot


def print_reaction_report(output_dir, rxn2annot):
    with open(output_dir + "/reactions.txt", "w") as rxn_fh:
        print("# Reaction frame ID", file=rxn_fh)
        print('\t'.join(["# ",
                         "Protein ID",
                         "Link Type",
                         "NOG",
                         "Rank",
                         "Taxon ID"]),
                        file=rxn_fh)
        for rxn in rxn2annot:
            print(rxn, file=rxn_fh)
            for (reaction_tuples, mapped_rank, mapped_taxon, prot_id, og) in rxn2annot[rxn]:
                print("\t".join(["",
                                 prot_id,
                                 [ rxn_tuple[5] for rxn_tuple in reaction_tuples if rxn_tuple[3] == rxn ][0],
                                 og,
                                 mapped_rank,
                                 mapped_taxon.tax_id]),
                      file=rxn_fh)

                
def print_multi_reaction_protein_report(output_dir, max_rxn_prot_annots):
    if len(max_rxn_prot_annots) > 0:
        with open (output_dir + "/multi-reaction-proteins.txt", "w") as multi_rxn_fh:
            print('\t'.join(["# Protein ID",
                             "NOG",
                             "Rank",
                             "Taxon ID",
                             "Reactions"]),
                  file=multi_rxn_fh)
            for (reaction_tuples, mapped_rank, mapped_taxon, prot_id, og) in max_rxn_prot_annots:
                print('\t'.join([prot_id,
                                 og,
                                 mapped_rank,
                                 mapped_taxon.tax_id,
                                 ';'.join([rxn_tuple[3] for rxn_tuple in reaction_tuples])]),
                      file=multi_rxn_fh)
    


def predict_pathways(eggnog_rxn_annot_file, ncbi_taxon_tuple, output_dir):

    rxn_ids = defaultdict(int)
    
    ## Collect reaction IDs:
    with open(eggnog_rxn_annot_file) as annot_fh:
        for line in annot_fh:
            if not line.startswith('#'):
                fields = line.strip().split('\t')
                if len(fields) == (rxn_field+1) and len(fields[rxn_field]) > 0:
                    curr_rxn_ids = fields[rxn_field].split(',')
                    for curr_rxn in curr_rxn_ids:
                        rxn_ids[curr_rxn]
    
    ## Send off to warpath:
    pwy_predictions = predict_metacyc_pathways(list(rxn_ids.keys()),
                                               ncbi_taxon_tuple)
    
    ## Print pathway predictions:
    #output_dir = os.path.dirname(os.path.abspath(eggnog_rxn_annot_file))
    report_file = output_dir + '/pathways.txt'
    with open(report_file, 'w') as pwy_fh:
        print('\t'.join(["# Pathway Frame ID",
                         "Pathway Name",
                         "Enzyme-Info-Content-Norm",
                         "Taxonomic Domain(s)",
                         "Present Reaction(s)",
                         "Missing Reaction(s)",
                         "All Key Reactions Present?",
                         "Key Reactions",
                         "Key Non Reactions Satisfied?",
                         "Key Non Reactions"]),
              file=pwy_fh)
        for pred in pwy_predictions:
            print('\t'.join([pred[0],
                             pred[1],
                             str("%0.3f" % pred[2]), 
                             ';'.join(pred[3]),
                             ';'.join(pred[4]),
                             ';'.join(pred[5]),
                             str(pred[6]),
                             ';'.join(pred[7]),
                             str(pred[8]),
                             ';'.join(pred[9])]),
                  file=pwy_fh)

    return (pwy_predictions, report_file)

def gen_metacyc_link(metacyc_frame):
    return 'https://metacyc.org/META/NEW-IMAGE?object=' + metacyc_frame

def gen_ncbi_tax_link(taxon_id):
    return 'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?mode=Info&id=' + \
        str(taxon_id) + \
        '&lvl=3&lin=f&keep=1&srchmode=1&unlock'

def gen_eggnog_link(nog):
    return 'http://eggnogdb.embl.de/#/app/results?target_nogs=' + nog.split('@')[0]


def gen_sphinx_hyperlink(link, text):
    return '`' + text + ' <' + link + '>`_'

def generate_ReST_report(output_dir, rxn2annot, pwy_pred_results):

    ncbi = NCBITaxa(reactionary.reactionary_lib.taxon_db_path)
    Path(output_dir + '/report-src').mkdir(parents=True, exist_ok=True)
    
    report_file = output_dir + '/report-src/index.rst'
    
    with open(report_file, 'w') as report_fh:

        ## Print the header:
        print(textwrap.dedent("""\
        Reactionary Report
        ==================
        
        This is a report generated by Reactionary (https://bitbucket.org/tomeraltman/reactionary/).
        
        .. contents:: Table of Contents
           :backlinks: entry

        Reactions
        ---------
        """),
              file = report_fh)

        ## Generate Reaction Table:
        
        print(".. csv-table:: Predicted Reactions", file = report_fh)
        print('   :header: "Reaction ID", "Protein ID", "Link Type", "NOG", "Rank", "Taxon ID", "Taxon Name"',
              file = report_fh)
        print('   :delim: |\n', file = report_fh)
        for rxn in rxn2annot:
            rxn_link = gen_sphinx_hyperlink(gen_metacyc_link(rxn), rxn)
            for (reaction_tuples, mapped_rank, mapped_taxon, prot_id, og) in rxn2annot[rxn]:
                taxon_link = gen_sphinx_hyperlink(gen_ncbi_tax_link(mapped_taxon.tax_id), mapped_taxon.tax_id)
                
                print("   " + "|".join([ rxn_link,
                                         prot_id,
                                         [ rxn_tuple[5] for rxn_tuple in reaction_tuples if rxn_tuple[3] == rxn ][0],
                                         gen_sphinx_hyperlink(gen_eggnog_link(og),og),
                                         mapped_rank,
                                         taxon_link,
                                         mapped_taxon.scientific_name]),
                      file=report_fh)
        print("", file = report_fh)

        
        ## Print Pathways section:
        print(textwrap.dedent("""\
        Pathways
        --------
        """),
              file = report_fh)
        
        ## Generate Pathway Table:

        print('.. csv-table:: Predicted Pathways', file = report_fh)
        print('   :header: ' + ', '.join(['"Pathway Frame ID"',
                                          '"Pathway Name"',
                                          '"Enzyme-Info-Content-Norm"',
                                          '"Taxonomic Domain(s)"',
                                          '"Present Reaction(s)"',
                                          '"Missing Reaction(s)"',
                                          '"All Key Reactions Present?"',
                                          '"Key Reactions"',
                                          '"Key Non Reactions Satisfied?"',
                                          '"Key Non Reactions"']),
              file = report_fh)
        print('   :delim: |\n', file = report_fh)
        for pred in pwy_pred_results:
            pwy_link = gen_sphinx_hyperlink(gen_metacyc_link(pred[0]), pred[0])
            taxonomic_domain_list = []
            for tax_id, name in ncbi.get_taxid_translator(pred[3]).items():
                rank = ncbi.get_rank([tax_id])[tax_id]
                taxonomic_domain_list.append(gen_sphinx_hyperlink(gen_ncbi_tax_link(tax_id),
                                                                  name + ' (' + rank + ')'))
                present_rxns = '; '.join([gen_sphinx_hyperlink(gen_metacyc_link(rxn),rxn)
                                          for rxn in pred[4] ])
                missing_rxns = '; '.join([gen_sphinx_hyperlink(gen_metacyc_link(rxn),rxn)
                                          for rxn in pred[5] ])
                key_rxns     = '; '.join([gen_sphinx_hyperlink(gen_metacyc_link(rxn),rxn)
                                          for rxn in pred[7] ])
                key_non_rxns = []
                if len(pred[9]) > 0:
                    key_non_rxns = '; '.join([pred[9][0]] + \
                                             [gen_sphinx_hyperlink(gen_metacyc_link(rxn),rxn)
                                              for rxn in pred[9][1:] ])
            print("   " + "|".join([ pwy_link,
                                     pred[1],
                                     str("%0.3f" % pred[2]), 
                                     '; '.join(taxonomic_domain_list),
                                     present_rxns,
                                     missing_rxns,
                                     str(pred[6]),
                                     key_rxns,
                                     str(pred[8]),
                                     key_non_rxns if len(key_non_rxns) > 0 else '\ ']),
                  file=report_fh)
        print("", file = report_fh)


    ## Run Sphinx to generate the report:
    sphinx_conf_file = os.path.dirname(__file__)
    output = subprocess.check_output(' '.join(['sphinx-build',
                                               '-b',
                                               'html',
                                               '-c',
                                               sphinx_conf_file,
                                               output_dir + '/report-src',
                                               output_dir]),
                                     shell=True)
        
              
            
### Deprecated

def coerce_taxonid_to_metacyc_taxon_frame(taxon_id_str,
                                          metaKB,
                                          email_address):
    
    Entrez.email = email_address
    
    ## Try to find taxon in MetaCyc; if successful, return the frame:
    ## "didn't find ID in MetaCyc, searching NCBI..."
    metacyc_taxon_frame = get_frame(metaKB,
                                    'TAX-' + taxon_id_str)
    
    if metacyc_taxon_frame != None:
        return metacyc_taxon_frame
    else:
        print("Didn't find given NCBI Taxonomy DB ID in MetaCyc, attempting to coerce using NCBI Entrez...", file=sys.stderr)
        
        ## If not, try to fetch the ID from NCBI Entrez:        
        try:
            handle = Entrez.efetch(db="taxonomy",
                                   id=taxon_id_str)
        
        ## Throw if not found:
        except:
            raise ValueError("NCBI Entrez did not recognize the given Taxon ID: '" + taxon_id_str + "'", taxon_id_str)
        
        ## Parse the XML:
        taxon_record = Entrez.read(handle)
        handle.close()
    
        ## Going to assume that list of hashes in XML is in the same order as the lineage,
        ## and thus will walk backwards until we find a hit:
        taxon_record[0]['LineageEx'].reverse()
        lineage = taxon_record[0]['LineageEx']
        for taxon_dict in lineage:
            metacyc_taxon_frame = get_frame(metaKB,
                                            'TAX-' + taxon_dict['TaxId'])
            if metacyc_taxon_frame != None:
                print("Reactionary: Coerced provided NCBI Taxonomy DB ID from '" + taxon_id_str + "' to '" + taxon_dict['TaxId'] + "' (" + taxon_dict['ScientificName'] + ", rank: '" + taxon_dict['Rank'] + "')", file=sys.stderr)
                break
    
    return metacyc_taxon_frame
        


    

    

### When run as a script:
#if __name__ == "__main__":



from __future__ import print_function
from Bio import Entrez
from ete3 import NCBITaxa
from camelot_frs.camelot_frs import get_kb, get_frame, get_frames, get_frame_all_parents, get_frame_all_children, frame_parent_of_frame_p, frame_subsumes_or_equal_p
from camelot_frs.pgdb_loader import load_pgdb
from camelot_frs.pgdb_api import uniprot_links_of_reaction, potential_generic_reaction_p, get_generic_reaction_all_subs, sub_reactions_of_reaction, enzymes_of_reaction, potential_pathway_hole_p

from collections import defaultdict, namedtuple
from random import shuffle
import sys, gzip, sqlite3

### Local definitions:

## Named tuples for clearer data-structure accessors:

Taxon = namedtuple('Taxon', ['tax_id', 'scientific_name', 'parent_tax_id', 'rank', 'lineage'])
Annot = namedtuple('Annot', ['prot_id', 'description', 'ecs', 'eggnog_ids'])

eggnog_mapper_ogg_field = 18

## This needs to be set by the code which imports this module:
taxon_db_path = ''

### Code for fetching NCBI Taxonomy taxon information:

## We need to use BioPython to take all of the EggNOG OG taxa, and
## fetch the "lineage" information along with ranks, to allow filtering:
# def fetch_taxon_info(tax_ids):
#     taxon2tuple = {}
    
#     epost_keys = Entrez.read(Entrez.epost('taxonomy', id=','.join(tax_ids)))
#     fetch_handle = Entrez.efetch(db='taxonomy',
#                                  webenv=epost_keys['WebEnv'],
#                                  query_key=epost_keys['QueryKey'],
#                                  retmode="xml")
#     taxa = Entrez.read(fetch_handle)
#     for taxon in taxa:
#         taxon_id = taxon['TaxId']
#         if 'LineageEx' in taxon:
#             curr_taxon = Taxon(tax_id = taxon_id,
#                                scientific_name = taxon['ScientificName'],
#                                parent_tax_id = taxon['ParentTaxId'],
#                                rank = taxon['Rank'],
#                                lineage = taxon['LineageEx'])
#         else:
#             curr_taxon = Taxon(tax_id = taxon['TaxId'],
#                                scientific_name = taxon['ScientificName'],
#                                parent_tax_id = taxon['ParentTaxId'],
#                                rank = taxon['Rank'],
#                                lineage = [])
#         taxon2tuple[taxon_id] = curr_taxon
    
#     return taxon2tuple

## using ete3, we report the data in the same format as Entrez returns:
def fetch_taxon_info(tax_ids):
    ncbi = NCBITaxa(taxon_db_path)
    taxon2tuple = {}
    for tax_id in tax_ids:                                        
        ## Remove the 'root' node from the beginning, and the tax_id from the end (redundant):
        curr_lineage = ncbi.get_lineage(tax_id)[1:-1]
        if curr_lineage:
            lineage_list = [ {'TaxId': str(lineage_taxon_id),
                              'ScientificName': ncbi.translate_to_names([lineage_taxon_id])[0],
                              'Rank': ncbi.get_rank([lineage_taxon_id])[lineage_taxon_id] }
                             for lineage_taxon_id in curr_lineage ]
            parent_id = str(curr_lineage[-1])
        else:
            lineage_list = []
            parent_id = '1'
        curr_rank = ncbi.get_rank([tax_id])
        curr_taxon = Taxon(tax_id = tax_id,
                           scientific_name = ncbi.translate_to_names([int(tax_id)])[0],
                           parent_tax_id = parent_id,
                           rank = curr_rank[int(tax_id)] if len(curr_rank) > 1 else 'no rank',
                           lineage = lineage_list)
        taxon2tuple[tax_id] = curr_taxon

    return taxon2tuple

## Helper predicate function
def ncbi_taxon_p(taxon_id):
    ncbi = NCBITaxa(taxon_db_path)
    return len(ncbi.get_rank([taxon_id])) > 0


## ToDo: figure out best way to serialize: JSON?
def print_taxon_file(taxon2lineage):
    pass


### Annotation file loading:
def load_annot_file(annot_file):
    annot_dict = {}
    with open(annot_file,'r') as annot_fh:
        for line in annot_fh:
            curr_annot = parse_annot_line(line)
            if curr_annot:
                annot_dict[curr_annot.prot_id] = curr_annot
    return annot_dict


def parse_annot_line (line):
    if not line.startswith('#'):
        fields = line.split("\t")
        fields[len(fields)-1] = fields[len(fields)-1].rstrip('\r\n')
        if fields[0].startswith('gnl'):
            prot_id = fields[0].split('|')[2]
        else:
            prot_id = fields[0]
        ecs = fields[7].split(',')
        eggnog_ids = fields[18].split(',')
        curr_annot = Annot(prot_id = prot_id,
                           description = fields[21],
                           ecs = ecs,
                           eggnog_ids = eggnog_ids)
        return curr_annot


### EggNOG mapping file:

RxnMap = namedtuple('RxnMap', ['OG',
                               'tax_id',
                               'uniprot_acc',
                               'rxn',
                               'enzrxn',
                               'relation',
                               'related_rxn',
                               'rhea_id'],
                    defaults=[None])
                            

def load_eggnog_rxn_mapping_file(mapping_file):
    
    eggnog2rxns = defaultdict(list)
    #eggnog2map  = defaultdict(list)
    #rxn2map     = defaultdict(list)
    #og2taxid    = {}
    
    with gzip.open(mapping_file,'rt') as nog2rxn_fh:
        for line in nog2rxn_fh:
            fields = line.strip().split('\t')
            if len(fields) == 6:
                curr_rxn_map = RxnMap._make(fields + [None])
            else:
                curr_rxn_map = RxnMap._make(fields)
                
            eggnog2rxns[fields[0]].append(curr_rxn_map)
            # eggnog2map[fields[0]].append(fields)
            # og2taxid[fields[0]] = fields[1]
            # if fields[5] == 'direct':
            #     eggnog2rxns[fields[0]].add(fields[3])
            #     rxn2map[fields[3]].append(fields)
            # else:
            #     eggnog2rxns[fields[0]].add(fields[6])
            #     rxn2map[fields[6]].append(fields)
    
    return eggnog2rxns


        
## Load MetaCyc effective orphan reactions:
def load_metacyc_effective_orphan_reactions(orphan_file):
    metacyc_orphan_rxns = defaultdict(int)
    with open(orphan_file) as orphan_fh:
        for line in orphan_fh:
            metacyc_orphan_rxns[line.strip()]
    return metacyc_orphan_rxns

### Load in EggNOG annotation of proteins:

## Print warnings about unrecognized taxon IDs from annotation file:
def fetch_taxa_of_annotation(annot_file):
    tax_ids = []
    present_tax_ids = []
    with open(annot_file) as annot_fh:
        for line in annot_fh:
            if not line.startswith('#'):
                fields = line.strip().split('\t')
                ogs = fields[18].split(',')
                tax_ids.extend([ og.split('@')[1] for og in ogs ])
    for tax_id in set(tax_ids):
        if not ncbi_taxon_p(tax_id):
            print('Warning: Unable to find the following annotation file taxon ID in local NCBI Taxonomy DB (will be skippped): ' + str(tax_id), file = sys.stderr)
        else:
            present_tax_ids.append(tax_id)
            #    tax_ids = list(set(tax_ids))
    return fetch_taxon_info(present_tax_ids)


## There are a couple of ways that we can pick an annotation:
## 1. Use the "bestOG"
## 2. Pick the most specific OG(s)
## 3. Pick the most specific OG(s) subject to a taxonomic constraint
##    (i.e., don't use an archaeal OG if the target genome is a gammaproteobacterium)
def load_eggnog_annotation(annot_file, nog2rxn_conn, annot_mode='most_specific'):
    num_annots = 0
    annot_go_terms  = defaultdict(int)
    annot_ogs       = defaultdict(int)
    annot_taxon2tuple = fetch_taxa_of_annotation(annot_file)
    
    with open(annot_file) as annot_fh:
        for line in annot_fh:
            if not line.startswith('#'):
                fields   = line.strip().split('\t')
                for go_term in fields[6].split(','):
                    annot_go_terms[go_term.strip()]
                best_og = find_og_in_annot_line(fields[18],"", nog2rxn_conn, annot_taxon2tuple, annot_mode=annot_mode)
                if best_og != "":
                    annot_ogs[best_og]
    return annot_ogs, annot_go_terms


def find_og_in_annot_struct(annot_struct,
                            nog2rxn_conn,
                            annot_taxon2tuple,
                            annot_mode='most_specific'):
    return find_og_in_annot_line(','.join(annot_struct.eggnog_ids),
                                 '',
                                 nog2rxn_conn,
                                 annot_taxon2tuple,
                                 annot_mode='most_specific')


def find_og_in_annot_line(ogs_field, best_og_field, nog2rxn_conn, annot_taxon2tuple, annot_mode='most_specific'):
    if annot_mode == 'most_specific':
        return select_most_specific_annotation(ogs_field.split(','), nog2rxn_conn, annot_taxon2tuple)[0]
    elif annot_mode == 'most_specific_v2':
        return select_most_specific_annotation_v2(ogs_field.split(','),
                                                  best_og_field.split('|')[0],
                                                  nog2rxn_conn)
    elif annot_mode == 'all':
        return load_all_annotations(ogs_field.split(','))
    elif annot_mode == 'best':
        return 'ENOG41'+ best_og_field.split('|')[0]

def ogs_in_db(conn, OG):
    cursor = conn.cursor()
    cursor.execute('select * from nog2rxn where OG = ?', (OG,))
    return cursor.fetchall()

## Returns the most specific OG which has a Reactionary annotation
## No taxonomic filtering takes place here, though it possibly
## would be a good point at which to do so.
def select_most_specific_annotation(raw_ogs, nog2rxn_conn, annot_taxon2tuple):
    most_specific_og = ""
    most_specific_og_taxon = ""
    most_specific_og_dist = 0
    for og in raw_ogs:
        ##og_id = og.split('@')[0]
        og_taxon = og.split('@')[1]
        if ncbi_taxon_p(og_taxon) and len(ogs_in_db(nog2rxn_conn, og)) > 0:
            if og_taxon in annot_taxon2tuple:
                og_dist  = len(annot_taxon2tuple[og_taxon].lineage)
            else:
                og_dist = 0
            if og_dist > most_specific_og_dist:
                most_specific_og_dist = og_dist
                most_specific_og = og
                most_specific_og_taxon = og_taxon
    return [most_specific_og, most_specific_og_taxon]

def ncbi_taxon_subsumes_or_equal_p(parent, child):
    if parent.tax_id == child.tax_id:
        return True
    else:
        parent_taxon = [ taxon for taxon in child.lineage if taxon['TaxId'] == parent.tax_id ]
        return len(parent_taxon) > 0

    
def ncbi_taxon_id_in_lineage_p(taxon_ids, ncbi_taxon):
    found = False
    for taxon_id in taxon_ids:
        for taxon in ncbi_taxon.lineage:
            if taxon['TaxId'] == taxon_id:
                found = True
    return found
    
def map_filter_taxon2domain(org_taxon):
    bacteria = Taxon(tax_id='2',
                     scientific_name='Bacteria',
                     parent_tax_id='131567',
                     rank='superkingdom',
                     lineage=[{'TaxId': '131567',
                               'ScientificName': 'cellular organisms',
                               'Rank': 'no rank'}])
    archaea = Taxon(tax_id='2157',
                    scientific_name='Archaea',
                    parent_tax_id='131567',
                    rank='superkingdom',
                    lineage=[{'TaxId': '131567',
                              'ScientificName': 'cellular organisms',
                              'Rank': 'no rank'}])
    eukaryota = Taxon(tax_id='2759',
                      scientific_name='Eukaryota',
                      parent_tax_id='131567',
                      rank='superkingdom',
                      lineage=[{'TaxId': '131567',
                                'ScientificName': 'cellular organisms',
                                'Rank': 'no rank'}])
    if ncbi_taxon_subsumes_or_equal_p(archaea, org_taxon):
        return archaea
    if ncbi_taxon_subsumes_or_equal_p(bacteria, org_taxon):
        return bacteria
    if ncbi_taxon_subsumes_or_equal_p(eukaryota, org_taxon):
        return eukaryota

    
def map_rank(rank, org_taxon, annot_taxon2tuple):
    
    rank_order = ['superkingdom',
                  'phylum',
                  'class',
                  'order',
                  'family',
                  'genus' ]
    
    
    ## First, try to map the rank to something more specific:
    given_rank_idx = rank_order.index(rank)
    for rank_idx in range(given_rank_idx, len(rank_order)):
        curr_rank = rank_order[rank_idx]
        target_tax_id = [taxon['TaxId'] for taxon in org_taxon.lineage
                         if taxon['Rank'] == curr_rank ][0]
        if target_tax_id in annot_taxon2tuple:
            return (curr_rank, annot_taxon2tuple[target_tax_id])
    
    ## Second, try to map the rank to something subsuming:
    given_rank_idx = rank_order.index(rank)
    for rank_idx in range(given_rank_idx, 0, -1):
        curr_rank = rank_order[rank_idx]
        target_tax_id = [taxon['TaxId'] for taxon in org_taxon.lineage
                         if taxon['Rank'] == curr_rank ][0]
        if target_tax_id in annot_taxon2tuple:
            return (curr_rank, annot_taxon2tuple[target_tax_id])
        

    
## Reaction Prediction

def mapOG2rxn_tuples(og, nog2rxn_conn, taxon2tuple, org_taxon, tax_restrict=True):
    reactions = []
    og_tax_id = og.split('@')[1]
    if tax_restrict and \
       og_tax_id in taxon2tuple and \
       ncbi_taxon_subsumes_or_equal_p(org_taxon, taxon2tuple[og_tax_id]):
        reactions = ogs_in_db(nog2rxn_conn, og)
    
    return reactions


'''
There are many ways to map from a list of OGs to one or more reactions.

This function takes a policy argument, and then applies it to the raw data.

* instance_rxns_p:
  boolean, indicates whether to return instance reactions as return values 
  if their subsuming generic reaction is also in the return set.

* max_rxns:
  Integer, max number of reactions to return. If an OG is so generic that it would pull in more than max_rxns number of reactions, then it is deemed uninformative and nothing is pulled in.
'''

def eggnog_annot2rxns(annot_struct,
                      org_taxon,
                      nog2rxn_conn,
                      annot_taxon2tuple,
                      instance_rxns_p=False,
                      enclosing_rank='superkingdom'):
    
    og = find_og_in_annot_struct(annot_struct,
                                 nog2rxn_conn,
                                 annot_taxon2tuple,
                                 annot_mode='most_specific')
    if not og:
        return [], '', 'no OG found with associated reactions', '', ''

    
    #filter_taxon = map_filter_taxon2domain(org_taxon)

    mapped_rank, mapped_taxon = map_rank(enclosing_rank,
                                         org_taxon,
                                         annot_taxon2tuple)
    #if mapped_rank != enclosing_rank:
    #    print("Warning: Reactionary taxonomic rank adjusted to: " + mapped_rank, file=sys.stderr)
        
    # filter_taxon_id = [taxon['TaxId'] for taxon in org_taxon.lineage
    #                    if taxon['Rank'] == enclosing_rank ][0]
    #
    #filter_taxon = annot_taxon2tuple[ mapped_tax_id ]
    
    reaction_tuples = mapOG2rxn_tuples(og,
                                       nog2rxn_conn,
                                       annot_taxon2tuple,                          
                                       mapped_taxon,
                                       tax_restrict=True)
    
    ## Remove instance reactions when subsuming generic reaction is present:
    ## No-op for now, premature optimization
    
    return (reaction_tuples,
            mapped_rank,
            mapped_taxon,
            annot_struct.prot_id,
            og)
                
    




#### Obsolete / Deprecated:

def load_all_annotations(raw_ogs):
    og_taxlevel_dict = {}
    print(raw_ogs)
    for og in raw_ogs:
        og_id = og.split('@')[0]
        og_level = og.split('@')[1]
        if og_id.startswith('COG'):                    
            og_taxlevel_dict['NOG'] = og_id
        else:                    
            og_taxlevel_dict[og_level] = 'ENOG41' + og_id
    return list(og_taxlevel_dict.values())

## This is a hack, until we can do proper taxonomic queries:
    # og_taxlevel_dict = {}
    # for og in raw_ogs:
    #     og_id = og.split('@')[0]
    #     og_level = og.split('@')[1]
    #     if og_id.startswith('COG'):                    
    #         og_taxlevel_dict['NOG'] = og_id
    #     else:                    
    #         og_taxlevel_dict[og_level] = 'ENOG41' + og_id
    # ## If there's something more specific, pick that one
    # ##        otherwise, first pick pronog, then bactnog, and lastly use NOG.
    # if 'bactNOG' in og_taxlevel_dict.keys():
    #     return og_taxlevel_dict['bactNOG']
    # if 'NOG' in og_taxlevel_dict.keys():
    #     return og_taxlevel_dict['NOG']
    # for taxlevel in og_taxlevel_dict.keys():
    #     if taxlevel not in ['NOG','bactNOG','proNOG']:
    #         return og_taxlevel_dict[taxlevel]
    # if 'proNOG' in og_taxlevel_dict.keys():
    #     return og_taxlevel_dict['proNOG']
    # return None

def select_most_specific_annotation_v2(raw_ogs,best_og,eggnog2rxns):
    ## First, see if the 'best' hit is in our annotation set:
    if best_og in eggnog2rxns:
        return best_og
    
    ## Try to find a specific NOG that has a hit in our annotationset:
    for og in raw_ogs:
        og_id = og.split('@')[0]
        og_level = og.split('@')[1]
        if not og_id.startswith('COG') and og_level not in ['bactNOG', 'euNOG', 'arNOG'] and ('ENOG41'+og_id) in eggnog2rxns:
            return 'ENOG41'+og_id
    
    ## If no specific hits, try to find a kingdom-level hit:
    for og in raw_ogs:
        og_id = og.split('@')[0]
        og_level = og.split('@')[1]
        if not og_id.startswith('COG') and og_level in ['bactNOG', 'euNOG', 'arNOG'] and ('ENOG41'+og_id) in eggnog2rxns:
            return 'ENOG41'+og_id
    
    ## If all else fails, try to find a COG-level hit:
    for og in raw_ogs:
        og_id = og.split('@')[0]
        og_level = og.split('@')[1]
        if og_id.startswith('COG') and og_id in eggnog2rxns:
            return og_id
    
    ## If nothing has a hit in the eggnog2rxns hash, then just return the 'best' hit
    return best_og

## Report how many "hops" from taxon to root:
def tax_dist2root(metaKB, tax_id):
    tax_frame = get_frame(metaKB, "TAX-" + tax_id)
    if tax_frame:
        return len(get_frame_all_parents(tax_frame)) - 2
    else:
        return 0







from setuptools import setup

setup(
    name='reactionary',
    version='0.4.1',
    description='Accurately annotate (meta)genomes with MetaCyc reactions.',
    author='Tomer Altman',
    author_email='python@me.tomeraltman.net',
    packages=['reactionary'],
    install_requires=['camelot-frs', 'biopython', 'ete3', 'docopt','schema', 'future', 'sphinx', 'pypandoc', 'rinohtype'], ##,'scikit-learn','scikit-plot','tkinter'],
    scripts=['bin/run-reactionary'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: POSIX",
        ],
    url='https://bitbucket.org/tomeraltman/reactionary/',
    include_package_data=True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import range
from past.utils import old_div
from camelot_frs.pgdb_loader import load_pgdb
from camelot_frs.camelot_frs import get_kb, get_frame, get_frames, get_frame_all_children
from .reactionary.reactionary_lib import load_eggnog_rxn_mapping_file, mapOG2rxn, find_og_in_annot_line
from .reactionary.warpath import predict_metacyc_pathways
from .reactionary.reactionary import coerce_taxonid_to_metacyc_taxon_frame, augment_eggnog_annotation
from Bio import Entrez
from collections import defaultdict
import sys, gzip, os
import matplotlib.pyplot as plt
import numpy as np

def eval_warpath(eggnog_rxn_annot_file, tax_id, metaKB):
    
    rxn_ids = defaultdict(int)
    
    ## Collect reaction IDs:
    with open(eggnog_rxn_annot_file) as annot_fh:
        for line in annot_fh:
            if not line.startswith('#'):
                fields = line.strip().split('\t')
                if len(fields) == (rxn_field+1) and len(fields[rxn_field]) > 0:
                    curr_rxn_ids = fields[rxn_field].split(',')
                    for curr_rxn in curr_rxn_ids:
                        rxn_ids[curr_rxn]
    
    ## Send off to warpath:
    pwy_predictions = predict_metacyc_pathways(get_frames(metaKB, list(rxn_ids.keys())),
                                               tax_id,
                                               metaKB)
    return pwy_predictions


rxn_field              = 22

eggnog_annot_file = os.path.dirname(__file__) + '/../test/ecocyc-annots/annot.emapper.annotations'

eggnog_rxn_annot_file = os.path.dirname(__file__) + '/../test/ecocyc-annots/annot.emapper.annotations.reactionary'

load_pgdb('/home/ubuntu/data/metacyc/23.1/data')
metaKB = get_kb('META')

tax_id = get_frame(metaKB, 'TAX-511145')

TODO: load annot_taxon2tuple

## Run EggNOG-Mapper over EcoCyc proteins:
augment_eggnog_annotation(eggnog_annot_file,
                          annot_taxon2tuple,
                          metaKB,
                          tax_id)

## Run predictions over EcoCyc EggNOG-Mapper output:
pwy_predictions = eval_warpath(eggnog_rxn_annot_file,
                               tax_id,
                               metaKB)

## Compute the true Pathways present in EcoCyc:
load_pgdb('/home/ubuntu/data/ecocyc/23.1/data')
ecoliKB = get_kb('ECOLI')

ecoli_pwys = [ pwy.frame_id for pwy in get_frame_all_children(get_frame(ecoliKB,'Pathways'),
                                                              frame_types='instance') ]

## Compute the TPR & FPR:
## TPR = TP/True = TP/(TP+FN)
## FPR = FP/False = FP/(FP+TN)

TP = defaultdict(int)
FP = defaultdict(int)
TN = defaultdict(int)
FN = defaultdict(int)
TPR = defaultdict(int)
FPR = defaultdict(int)
possible_metacyc_pwys = defaultdict(int)
threshold = defaultdict(int)
acc = defaultdict(int)
F1_score = defaultdict(int)

num_samples = 40
for i in range(0,num_samples+1):
    threshold[i] = i*1.0/num_samples
    for pwy, rxn_frac, _ in pwy_predictions:
        possible_metacyc_pwys[pwy.frame_id]
        if pwy.frame_id in ecoli_pwys:
            if rxn_frac >= threshold[i]:            
                TP[i] += 1
            else:
                FN[i] += 1
        else:
            if rxn_frac >= threshold[i]:
                FP[i] += 1
            else:
                TN[i] += 1
    TPR[i] = TP[i]*1.0/(TP[i]+FN[i])
    FPR[i] = FP[i]*1.0/(FP[i]+TN[i])
    acc[i] = old_div(1.0*(TP[i]+TN[i]),(TP[i]+TN[i]+FP[i]+FN[i]))
    F1_score[i] = old_div(2.0*TP[i],(2*TP[i]+FP[i]+FN[i]))


print(list(TPR.values()))
print(list(FPR.values()))

## ROC Curve:
plt.plot(list(FPR.values())+[0], list(TPR.values())+[0])
plt.ylim(0,1.0)
plt.xlim(0,1.0)
plt.show()

## Accuracy vs. threshold:
plt.plot(list(threshold.values()),list(F1_score.values()))
plt.ylim(0,1.0)
plt.xlim(0,1.0)
plt.show()
auc=np.trapz(list(TPR.values())+[0],list(FPR.values())+[0])
            
    
    
